import platform
from pathlib import Path
from tempfile import TemporaryDirectory

import pandas as pd
import pytest

from advertools.spider import crawl

system = platform.system()

links_columns = {
    "links_url": 14,
    "links_text": 14,
    "links_nofollow": 14,
    "nav_links_url": 3,
    "nav_links_text": 3,
    "header_links_url": 3,
    "header_links_text": 3,
    "footer_links_url": 3,
    "footer_links_text": 3,
}

links_filepath = "tests/data/crawl_testing/test_content.html"
if platform == "Windows":
    links_filepath = links_filepath.replace("/", r"\\")
links_file = Path(links_filepath).absolute()

dup_links_file_path = "tests/data/crawl_testing/duplicate_links.html"
if platform == "Windows":
    dup_links_file_path = dup_links_file_path.replace("/", r"\\")
dup_links_file = Path(dup_links_file_path).absolute()

broken_links_path = "tests/data/crawl_testing/broken_links.html"
if platform == "Windows":
    broken_links_path = links_filepath.replace("/", r"\\")

broken_links_file = Path(broken_links_path).absolute()

with TemporaryDirectory() as links_crawl_tempdir:
    crawl(
        links_file.as_uri(),
        f"{links_crawl_tempdir}/links_crawl.jl",
        custom_settings={"ROBOTSTXT_OBEY": False},
    )
    crawl_df = pd.read_json(f"{links_crawl_tempdir}/links_crawl.jl", lines=True)

    def test_link_columns_all_exist():
        assert set(links_columns).difference(crawl_df.columns.tolist()) == set()

    @pytest.mark.parametrize("colname,count", links_columns.items())
    def test_links_extracted_at_correct_number(colname, count):
        assert crawl_df[colname].str.split("@@").str.len().values[0] == count

    def test_extract_h_tags():
        assert crawl_df["h2"].str.split("@@").str.len().values[0] == 3
        assert crawl_df["h2"].str.split("@@").explode().iloc[1] == ""

    def test_all_links_have_nofollow():
        assert (
            crawl_df.filter(regex="nofollow")
            .apply(lambda s: s.str.contains("True"))
            .all()
            .all()
        )

    def test_image_tags_available():
        assert [
            col in crawl_df for col in ["img_src", "img_alt", "img_height", "img_width"]
        ]

    def test_all_img_attrs_have_same_length():
        assert (
            crawl_df.filter(regex="img_")
            .apply(lambda s: s.str.split("@@").str.len())
            .apply(set, axis=1)[0]
            .__len__()
        ) == 1

    def test_img_src_has_abs_path():
        assert crawl_df["img_src"].str.startswith("http").all()


with TemporaryDirectory() as follow_url_params_tempdir:
    crawl(
        str(links_file.as_uri()),
        f"{follow_url_params_tempdir}/follow_url_params.jl",
        allowed_domains=["", "example.com"],
        custom_settings={"ROBOTSTXT_OBEY": False},
        follow_links=True,
    )

    follow_url_params_df = pd.read_json(
        f"{follow_url_params_tempdir}/follow_url_params.jl", lines=True
    )

    def test_follow_url_params_followed():
        assert follow_url_params_df["url"].str.contains("?", regex=False).any()


with TemporaryDirectory() as dont_follow_url_params_tempdir:
    crawl(
        str(links_file.as_uri()),
        f"{dont_follow_url_params_tempdir}/dont_follow_url_params.jl",
        allowed_domains=["", "example.com"],
        custom_settings={"ROBOTSTXT_OBEY": False},
        follow_links=True,
        exclude_url_params=True,
    )
    dont_follow_url_params_df = pd.read_json(
        f"{dont_follow_url_params_tempdir}/dont_follow_url_params.jl", lines=True
    )

    def test_dont_follow_url_params_not_followed():
        assert not dont_follow_url_params_df["url"].str.contains("?", regex=False).all()


with TemporaryDirectory() as dup_links_file_tempdir:
    crawl(
        str(dup_links_file.as_uri()),
        f"{dup_links_file_tempdir}/dup_links_crawl.jl",
        custom_settings={"ROBOTSTXT_OBEY": False},
    )
    dup_crawl_df = pd.read_json(
        f"{dup_links_file_tempdir}/dup_links_crawl.jl", lines=True
    )
    dup_links_test = ["https://example_a.com" for i in range(5)] + [
        "https://example.com"
    ]

    dup_text_test = [
        "Link Text A",
        "Link Text A",
        "Link Text A",
        "Link Text B",
        "Link Text C",
        "Link Other",
    ]

    dup_nf_test = ["True"] + ["False" for i in range(5)]

    def test_duplicate_links_counted_propery():
        assert dup_crawl_df["links_url"].str.split("@@")[0] == dup_links_test
        assert dup_crawl_df["links_text"].str.split("@@")[0] == dup_text_test
        assert dup_crawl_df["links_nofollow"].str.split("@@")[0] == dup_nf_test

    def test_non_existent_links_are_NA():
        assert "nav_links_url" not in dup_crawl_df
        assert "nav_links_text" not in dup_crawl_df
        assert "header_links_url" not in dup_crawl_df
        assert "footer_links_url" not in dup_crawl_df


with TemporaryDirectory() as broken_links_tempdir:

    def test_broken_links_are_reported():
        crawl(
            url_list=["https://wikipedia.org", "https://wrong_url"],
            output_file=f"{broken_links_tempdir}/broken_links_crawl.jl",
            custom_settings={"ROBOTSTXT_OBEY": False},
        )
        broken_links_df = pd.read_json(
            f"{broken_links_tempdir}/broken_links_crawl.jl", lines=True
        )
        assert "errors" in broken_links_df
        assert "https://wrong_url" in broken_links_df["url"].tolist()

    def test_crawling_bad_url_directly_is_handled():
        crawl(
            ["wrong_url", "https://example.com"], f"{broken_links_tempdir}/bad_url.jl"
        )
        bad_url_df = pd.read_json(f"{broken_links_tempdir}/bad_url.jl", lines=True)
        assert len(bad_url_df) == 1
        assert bad_url_df["url"][0] == "https://example.com"



from pathlib import Path

import pandas as pd
import pytest

from advertools.sitemaps import sitemap_to_df, _build_request_headers, headers as DEFAULT_HEADERS

gh_test_data_folder = "https://raw.githubusercontent.com/eliasdabbas/advertools/master/tests/data/sitemap_testing/"
offline_test_data_folder = "tests/data/sitemap_testing/"


def full_path(file):
    return gh_test_data_folder + file


def offline_path(filename):
    path = Path(offline_test_data_folder + filename)
    return path.absolute().as_uri()


regular_sitemap_url = full_path("regular_sitemap.xml")
zipped_sitemap_url = full_path("zipped_sitemap.xml.gz")
sitemap_index_url = full_path("sitemap_index.xml")
error_sitemap_url = full_path("error_sitemap.xml")
image_sitemap_url = full_path("image_sitemap.xml")
video_sitemap_url = full_path("video_sitemap.xml")
news_sitemap_url = full_path("news_sitemap.xml")
robotstxt_url = full_path("robots.txt")


def test_build_request_headers():
    user_headers = {"If-None-Match": "ETAG_STRING"}
    final_headers = _build_request_headers(user_headers)
    assert isinstance(final_headers, dict)
    assert final_headers == {
        "user-agent": DEFAULT_HEADERS["User-Agent"],
        "if-none-match": "ETAG_STRING"
    }


def test_build_request_headers_default():
    user_headers = None
    final_headers = _build_request_headers(user_headers)
    assert final_headers == {"user-agent": DEFAULT_HEADERS["User-Agent"]}


def test_build_request_headers_override_default():
    user_headers = {"User-agent": "example/agent", "If-None-Match": "ETAG_STRING"}
    final_headers = _build_request_headers(user_headers)
    assert final_headers == {
        "user-agent": "example/agent",
        "if-none-match": "ETAG_STRING"
    }


def test_regular_sitemap():
    result = sitemap_to_df(regular_sitemap_url)
    assert isinstance(result, pd.core.frame.DataFrame)
    assert len(result) == 5


def test_gz_sitemap():
    result = sitemap_to_df(zipped_sitemap_url)
    assert isinstance(result, pd.core.frame.DataFrame)
    assert len(result) == 5


def test_sitemap_index():
    result = sitemap_to_df(sitemap_index_url)
    assert isinstance(result, pd.core.frame.DataFrame)
    assert "errors" in result
    errors = {"WARNING: Sitemap contains a link to itself", "HTTP Error 404: Not Found"}
    assert errors.issubset(result["errors"])
    assert all([col in result for col in ["loc", "download_date", "sitemap"]])


def test_error_sitemap():
    with pytest.raises(Exception):
        sitemap_to_df(error_sitemap_url)


def test_image_sitemap():
    result = sitemap_to_df(image_sitemap_url)
    assert isinstance(result, pd.core.frame.DataFrame)
    assert "image" in result


def test_video_sitemap():
    result = sitemap_to_df(video_sitemap_url)
    assert isinstance(result, pd.core.frame.DataFrame)
    assert "video_content_loc" in result


def test_news_sitemap():
    result = sitemap_to_df(news_sitemap_url)
    assert isinstance(result, pd.core.frame.DataFrame)
    assert "news" in result


def test_get_sitemaps_from_robotstxt():
    result = sitemap_to_df(robotstxt_url)
    assert isinstance(result, pd.core.frame.DataFrame)


def test_sitemaps_offline():
    sitemap_path = offline_path("regular_sitemap.xml")
    result = sitemap_to_df(sitemap_path)
    assert isinstance(result, pd.DataFrame)


def test_zipped_sitemaps_offline():
    sitemap_path = offline_path("zipped_sitemap.xml.gz")
    result = sitemap_to_df(sitemap_path)
    assert isinstance(result, pd.DataFrame)



import os

import pandas as pd
import pytest

import advertools.youtube as yt

youtube_key = os.environ.get("GOOG_CSE_KEY")

skip_api_tests = pytest.mark.skipif(
    os.environ.get("ADV_TEST_OFFLINE"), reason="Run all except API dependents"
)


@skip_api_tests
def test_activities_list():
    result = yt.activities_list(
        key=youtube_key, part="snippet", channelId="UCv002AUCZaPNwiADqwchijg"
    )
    assert {"queryTime", "param_part"}.issubset(result.columns)


@skip_api_tests
def test_captions_list():
    result = yt.captions_list(key=youtube_key, part="snippet", videoId="kJQP7kiw5Fk")
    assert {"queryTime", "param_part"}.issubset(result.columns)


@skip_api_tests
def test_channel_sections_list():
    result = yt.channel_sections_list(
        key=youtube_key, part="snippet", channelId="UCv002AUCZaPNwiADqwchijg"
    )
    assert {"queryTime", "param_part"}.issubset(result.columns)


@skip_api_tests
def test_channels_list():
    result = yt.channels_list(
        key=youtube_key, part="snippet,statistics", forUsername="youtube"
    )
    assert {"queryTime", "param_part"}.issubset(result.columns)


@skip_api_tests
def test_comment_threads_list():
    result = yt.comment_threads_list(
        key=youtube_key, part="id,replies,snippet", videoId="PscrVidwxMg"
    )
    assert {"queryTime", "param_part"}.issubset(result.columns)


@skip_api_tests
def test_comments_list():
    result = yt.comments_list(
        key=youtube_key, part="snippet", id="UgxKMCc9z4iE7LNW2Hh4AaABAg"
    )
    assert {"queryTime", "param_part"}.issubset(result.columns)


# @skip_api_tests
# def test_guide_categories_list():
#     result = yt.guide_categories_list(key=youtube_key, part='snippet',
#                                       regionCode='tr', hl='tr')
#     assert {'queryTime', 'param_part'}.issubset(result.columns)


@skip_api_tests
def test_i18n_languages_list():
    result = yt.i18n_languages_list(key=youtube_key, part="snippet")
    assert {"queryTime", "param_part"}.issubset(result.columns)


@skip_api_tests
def test_i18n_regions_list():
    result = yt.i18n_regions_list(key=youtube_key, part="snippet")
    assert {"queryTime", "param_part"}.issubset(result.columns)


@skip_api_tests
def test_playlist_items_list():
    result = yt.playlist_items_list(
        key=youtube_key, part="snippet", playlistId="PLW0Gy9pTgVntoeYT50HfV144rzuJcrmMg"
    )
    assert {"queryTime", "param_part"}.issubset(result.columns)


@skip_api_tests
def test_playlists_list():
    result = yt.playlists_list(
        key=youtube_key, part="snippet", channelId="UCv002AUCZaPNwiADqwchijg"
    )
    assert {"queryTime", "param_part"}.issubset(result.columns)


@skip_api_tests
def test_search():
    result = yt.search(key=youtube_key, part="snippet", q="test bitcoin")
    assert {"queryTime", "param_part"}.issubset(result.columns)


@skip_api_tests
def test_subscriptions_list():
    result = yt.subscriptions_list(
        key=youtube_key, part="snippet", channelId="UCv002AUCZaPNwiADqwchijg"
    )
    assert {"queryTime", "param_part"}.issubset(result.columns)


@skip_api_tests
def test_video_categories_list():
    result = yt.video_categories_list(key=youtube_key, part="snippet", regionCode="de")
    assert {"queryTime", "param_part"}.issubset(result.columns)


@skip_api_tests
def test_videos_list():
    result = yt.videos_list(
        key=youtube_key,
        part="snippet",
        chart="mostPopular",
        regionCode="GB",
        maxResults=9,
    )
    assert {"queryTime", "param_part"}.issubset(result.columns)


# Test raising errors:


@skip_api_tests
def test_activities_list_raises():
    with pytest.raises(ValueError):
        yt.activities_list(key=youtube_key, part="wrong_part")

    with pytest.raises(ValueError):
        yt.activities_list(key=youtube_key, part="snippet")


@skip_api_tests
def test_captions_list_raises():
    with pytest.raises(ValueError):
        yt.captions_list(key=youtube_key, part="wrong_part", videoId="random")


@skip_api_tests
def test_channel_sections_list_raises():
    with pytest.raises(ValueError):
        yt.channel_sections_list(key=youtube_key, part="wrong_part")

    with pytest.raises(ValueError):
        yt.channel_sections_list(key=youtube_key, part="snippet")


@skip_api_tests
def test_channels_list_raises():
    with pytest.raises(ValueError):
        yt.channels_list(key=youtube_key, part="wrong_part")

    with pytest.raises(ValueError):
        yt.channels_list(key=youtube_key, part="snippet")


@skip_api_tests
def test_comment_threads_list_raises():
    with pytest.raises(ValueError):
        yt.comment_threads_list(key=youtube_key, part="wrong_part")

    with pytest.raises(ValueError):
        yt.comment_threads_list(key=youtube_key, part="snippet")


@skip_api_tests
def test_comments_list_raises():
    with pytest.raises(ValueError):
        yt.comments_list(key=youtube_key, part="wrong_part")

    with pytest.raises(ValueError):
        yt.comments_list(key=youtube_key, part="snippet")


@skip_api_tests
def test_guide_categories_list_raises():
    with pytest.raises(ValueError):
        yt.guide_categories_list(key=youtube_key, part="wrong_part")

    with pytest.raises(ValueError):
        yt.guide_categories_list(key=youtube_key, part="snippet")


@skip_api_tests
def test_playlist_items_list_raises():
    with pytest.raises(ValueError):
        yt.playlist_items_list(key=youtube_key, part="wrong_part")

    with pytest.raises(ValueError):
        yt.playlist_items_list(key=youtube_key, part="snippet")


@skip_api_tests
def test_playlists_list_raises():
    with pytest.raises(ValueError):
        yt.playlists_list(key=youtube_key, part="wrong_part")

    with pytest.raises(ValueError):
        yt.playlists_list(key=youtube_key, part="snippet")


@skip_api_tests
def test_subscriptions_list_raises():
    with pytest.raises(ValueError):
        yt.subscriptions_list(key=youtube_key, part="wrong_part")

    with pytest.raises(ValueError):
        yt.subscriptions_list(key=youtube_key, part="snippet")


@skip_api_tests
def test_video_categories_list_raises():
    with pytest.raises(ValueError):
        yt.video_categories_list(key=youtube_key, part="wrong_part")

    with pytest.raises(ValueError):
        yt.video_categories_list(key=youtube_key, part="snippet")


@skip_api_tests
def test_videos_list_raises():
    with pytest.raises(ValueError):
        yt.videos_list(key=youtube_key, part="wrong_part")

    with pytest.raises(ValueError):
        yt.videos_list(key=youtube_key, part="snippet")


@skip_api_tests
def test_with_zero_results_required():
    result = yt.search(
        key=youtube_key,
        part="snippet",
        q="testing bitcoin",
        maxResults=0,
    )
    assert isinstance(result, pd.DataFrame)


@skip_api_tests
def test_empty_list_raises_error():
    with pytest.raises(ValueError):
        yt.search(key=youtube_key, part="snippet", q="testing bitcoin", regionCode=[])


@skip_api_tests
def test_errors_returned_as_df():
    result = yt.search(key="wrong key", part="snippet", q="testing bitcoin")
    assert "errors" in result



import os
from pathlib import Path
from tempfile import TemporaryDirectory

import pandas as pd
import pytest

from advertools.logs import crawllogs_to_df, logs_to_df


def full_local_path(folder, file):
    tests_dir = f"tests/data/{folder}/"
    return os.path.abspath(tests_dir + file)


@pytest.mark.parametrize(
    "logfile",
    [f for f in os.listdir("tests/data/logs_testing") if f.startswith("crawl")],
)
def test_crawllogs_returns_df(logfile):
    result = crawllogs_to_df(full_local_path("logs_testing", logfile))
    assert isinstance(result, pd.core.frame.DataFrame)


def test_logstodf_raises_on_wrong_output_file():
    with pytest.raises(ValueError):
        logs_to_df("logfile.log", "output.wrong", "errors.txt", "common")


def test_logstodf_general_output_correctness():
    with TemporaryDirectory() as tmpdir:
        path = Path(tmpdir)
        logs_to_df(
            full_local_path("logs_testing", "nginx_access.log"),
            str(path / "delete_output.parquet"),
            path / "delete_errors.txt",
            "combined",
        )
        result = pd.read_parquet(path / "delete_output.parquet")
        assert result.filter(regex="referer|user_agent").columns.tolist() == [
            "referer",
            "user_agent",
        ]


def test_logs_raises_for_wrong_encoding():
    with pytest.raises(UnicodeDecodeError):
        logs_to_df(
            full_local_path("logs_testing", "combined_latin1_enc.log"),
            "test_output.parquet",
            "test_errors.txt",
            "combined",
            encoding="utf-8",
        )


def test_logs_parses_with_correct_encoding():
    with TemporaryDirectory() as tmpdir:
        path = Path(tmpdir)
        logs_to_df(
            full_local_path("logs_testing", "combined_latin1_enc.log"),
            str(path / "test_output.parquet"),
            path / "test_errors.txt",
            "combined",
            encoding="latin1",
        )
        result = pd.read_parquet(path / "test_output.parquet")
        assert isinstance(result, pd.DataFrame)


def test_logs_nginx_error():
    with TemporaryDirectory() as tmpdir:
        path = Path(tmpdir)
        logs_to_df(
            full_local_path("logs_testing", "nginx_error.log"),
            str(path / "test_output.parquet"),
            path / "test_errors.txt",
            "nginx_error",
            encoding="latin1",
        )
        result = pd.read_parquet(path / "test_output.parquet")
        assert isinstance(result, pd.DataFrame)


def test_logs_apache_error():
    with TemporaryDirectory() as tmpdir:
        path = Path(tmpdir)
        logs_to_df(
            full_local_path("logs_testing", "apache_error.log"),
            str(path / "test_output.parquet"),
            path / "test_errors.txt",
            "nginx_error",
            encoding="latin1",
        )
        result = pd.read_parquet(path / "test_output.parquet")
        assert isinstance(result, pd.DataFrame)


def test_logs_file_without_errors():
    with TemporaryDirectory() as tmpdir:
        path = Path(tmpdir)
        logs_to_df(
            full_local_path("logs_testing", "combined_no_errors.log"),
            str(path / "delete_output.parquet"),
            path / "delete_errors.txt",
            "combined",
        )
        assert not os.path.exists(path / "delete_errors.txt")


def test_logstodf_parses_dates():
    with TemporaryDirectory() as tmpdir:
        path = Path(tmpdir)
        logs_to_df(
            full_local_path("logs_testing", "nginx_access.log"),
            str(path / "delete_output.parquet"),
            path / "delete_errors.txt",
            "combined",
        )
        result = pd.read_parquet(path / "delete_output.parquet")
        assert "datetime" in result["datetime"].dtype.name


def test_logstodf_parses_custom_date_fmt():
    with TemporaryDirectory() as tmpdir:
        path = Path(tmpdir)
        logs_to_df(
            full_local_path("logs_testing", "nginx_custom_date.log"),
            str(path / "delete_output.parquet"),
            path / "delete_errors.txt",
            "combined",
            date_format="%B %d, %YT%H:%M:%S %z",
        )
        result = pd.read_parquet(path / "delete_output.parquet")
        assert "datetime" in result["datetime"].dtype.name


def test_logstodf_parses_raises_wrong_date_fmt():
    with TemporaryDirectory() as tmpdir:
        path = Path(tmpdir)
        logs_to_df(
            full_local_path("logs_testing", "nginx_custom_date.log"),
            str(path / "delete_output.parquet"),
            path / "delete_errors.txt",
            "combined",
            date_format="%B %, M:%S %z",
        )
        result = pd.read_parquet(path / "delete_output.parquet")
        assert not ("datetime" in result["datetime"].dtype.name)



from urllib.parse import parse_qs, urlparse

import pytest


from advertools.url_builders import url_utm_ga


def test_correct_url_returned():
    url = "https://www.mysite.com?utm_source=source&utm_medium=medium&utm_campaign=campaign&utm_content=content&utm_term=term"
    composed_url = url_utm_ga(
        "www.mysite.com", "source", "medium", "campaign", "content", "term"
    )

    parsed = urlparse(url)
    assert parse_qs(parsed.query) == parse_qs(urlparse(composed_url).query)


def test_raises_error_missing_source():
    with pytest.raises(TypeError):
        url_utm_ga("www.mysite.com")


def test_raises_error_missing_url():
    with pytest.raises(TypeError):
        url_utm_ga(source="source")


def test_space_converted_to_plut():
    result = url_utm_ga("www.mysite.com", "one two")
    assert result == "www.mysite.com?utm_source=one+two"


def test_symbols_are_url_encoded():
    result = url_utm_ga("www.mysite.com", "one !@#$%^&*() two")
    assert result == "www.mysite.com?utm_source=one+%21%40%23%24%25%5E%26%2A%28%29+two"



import pytest

from advertools.emoji import extract_emoji
from advertools.extract import (
    extract,
    extract_currency,
    extract_exclamations,
    extract_hashtags,
    extract_intense_words,
    extract_mentions,
    extract_numbers,
    extract_questions,
    extract_urls,
    extract_words,
)

number_posts = [
    "before123,000",
    "123after",
    "comma 123,456",
    "dot 123.234.3",
    "skip 123-",
    "nothing",
    "two 123 456,789",
]

number_summary = extract_numbers(number_posts)

number_test_keys = [
    "numbers",
    "numbers_flat",
    "number_counts",
    "number_freq",
    "top_numbers",
    "overview",
]

mention_posts = [
    "hello @name",
    "email@domain.com",
    "@oneword",
    "hi @nam-e and @name",
    "@first @last",
    "an @under_score",
    "a @dot.one",
    "non latin @مرحبا",
    "other at ＠sign",
    "@one.@two three",
    "number @123text",
    "@_before @after_",
]

mention_summary = extract_mentions(mention_posts)

mention_test_keys = [
    "mentions",
    "mentions_flat",
    "mention_counts",
    "mention_freq",
    "top_mentions",
    "overview",
]

hashtag_posts = [
    "hello #name",
    "email#domain.com",
    "#oneword",
    "hi #nam-e and #name",
    "#first #last",
    "an #under_score",
    "a #dot.one",
    "non latin #مرحبا",
    "other hash ＃sign",
    "#one.#two three",
    "number #123text",
    "#_before #after_",
]

hashtag_summary = extract_hashtags(hashtag_posts)

hashtag_test_keys = [
    "hashtags",
    "hashtags_flat",
    "hashtag_counts",
    "hashtag_freq",
    "top_hashtags",
    "overview",
]

emoji_posts = ["one smiley 😀", "one smiley 😀 one wink 😉", "no emoji"]

emoji_summary = extract_emoji(emoji_posts)

emoji_test_keys = [
    "emoji",
    "emoji_text",
    "emoji_flat",
    "emoji_flat_text",
    "emoji_counts",
    "emoji_freq",
    "top_emoji",
    "top_emoji_text",
    "top_emoji_groups",
    "top_emoji_sub_groups",
    "overview",
]

word_posts = [
    "today it is raining",
    "i like rain and raining",
    "today it is snowing",
    "now snowing and raining",
    "training is good with snow",
    "RaIn and SNoW",
    "none of the words",
    "@rain, #snow rain",
]

word_summary_full = extract_words(word_posts, ["rain", "snow"], True)

word_summary_not_full = extract_words(word_posts, ["rain", "snow"], False)

word_test_keys = [
    "words",
    "words_flat",
    "word_counts",
    "word_freq",
    "top_words",
    "overview",
]

currency_posts = [
    "$5.0 beginning",
    "mid £5.0 price",
    "end of str €",
    "more $ than £ one",
    "other symbols ₣ ₤ ₥ ₦ ₧",
    "no symbols",
]

currency_summary = extract_currency(currency_posts)

currency_test_keys = [
    "currency_symbols",
    "currency_symbols_flat",
    "currency_symbol_counts",
    "currency_symbol_freq",
    "top_currency_symbols",
    "overview",
    "currency_symbol_names",
    "surrounding_text",
]

intense_word_posts = [
    "i looooove this",
    "goooood mooorning",
    "normal text",
    "in thhhhhhe middle",
    "innnnnn the beginning",
    "at the end!!!!!!",
]

intense_word_summary = extract_intense_words(intense_word_posts)

intense_word_test_keys = [
    "intense_words",
    "intense_words_flat",
    "intense_word_counts",
    "intense_word_freq",
    "top_intense_words",
    "overview",
]

question_posts = [
    "how are you?",
    "no question",
    "no! what about you?",
    "Hola, ¿cómo estás?",
    "Πώς είσαι;",
]

question_summary = extract_questions(question_posts)

question_test_keys = [
    "question_marks",
    "question_marks_flat",
    "question_mark_counts",
    "question_mark_freq",
    "top_question_marks",
    "overview",
    "question_mark_names",
    "question_text",
]

exclamation_posts = [
    "how dare you!",
    "no exclamation",
    "no! do not do this!",
    "¡Hola!  ¿cómo estás?",
    "مرحبا. لا تذهب!",
]

exclamation_summary = extract_exclamations(exclamation_posts)

exclamation_test_keys = [
    "exclamation_marks",
    "exclamation_marks_flat",
    "exclamation_mark_counts",
    "exclamation_mark_freq",
    "top_exclamation_marks",
    "overview",
    "exclamation_mark_names",
    "exclamation_text",
]

url_posts = [
    "one https://www.a.com",
    "two www.a.com www.b.com",
    "nothing",
    "long https://example.com/one?a=b#nothing",
]

url_summary = extract_urls(url_posts)

url_test_keys = [
    "urls",
    "urls_flat",
    "url_counts",
    "url_freq",
    "top_urls",
    "overview",
    "top_domains",
    "top_tlds",
]

test_ids = [
    "currency",
    "emoji",
    "exclamation",
    "hashtag",
    "intense",
    "mention",
    "number",
    "question",
    "word_full",
    "word_not_full",
    "url",
]


summaries_keys = [
    (list(currency_summary.keys()), currency_test_keys),
    (list(emoji_summary.keys()), emoji_test_keys),
    (list(exclamation_summary.keys()), exclamation_test_keys),
    (list(hashtag_summary.keys()), hashtag_test_keys),
    (list(intense_word_summary.keys()), intense_word_test_keys),
    (list(mention_summary.keys()), mention_test_keys),
    (list(number_summary.keys()), number_test_keys),
    (list(question_summary.keys()), question_test_keys),
    (list(word_summary_full.keys()), word_test_keys),
    (list(word_summary_not_full.keys()), word_test_keys),
    (list(url_summary.keys()), url_test_keys),
]


@pytest.mark.parametrize(
    "summary_key, test_key",
    zip([x[0] for x in summaries_keys], [x[1] for x in summaries_keys]),
    ids=test_ids,
)
def test_has_correct_keys(summary_key, test_key):
    assert set(summary_key) == set(test_key)


summaries = {
    "currency": currency_summary,
    "emoji": emoji_summary,
    "exclamation": exclamation_summary,
    "hashtag": hashtag_summary,
    "intense": intense_word_summary,
    "mention": mention_summary,
    "number": number_summary,
    "question": question_summary,
    "word_full": word_summary_full,
    "word_not_full": word_summary_not_full,
    "url": url_summary,
}


summary_counts = {
    k: [d[key] for key in d if "count" in key][0] for k, d in summaries.items()
}
test_counts = {
    "currency": [1, 1, 1, 2, 5, 0],
    "emoji": [1, 2, 0],
    "exclamation": [1, 0, 2, 2, 1],
    "hashtag": [1, 0, 1, 2, 2, 1, 1, 1, 1, 2, 1, 2],
    "intense": [1, 2, 0, 1, 1, 1],
    "mention": [1, 0, 1, 2, 2, 1, 1, 0, 1, 2, 1, 2],
    "number": [1, 1, 1, 1, 10, 2],
    "question": [1, 0, 1, 2, 1],
    "word_full": [0, 1, 0, 0, 1, 2, 0, 3],
    "word_not_full": [1, 2, 1, 2, 2, 2, 0, 3],
    "url": [1, 2, 0, 1],
}


@pytest.mark.parametrize(
    "summary_count, test_count",
    zip(sorted(summary_counts.items()), sorted(test_counts.items())),
    ids=test_ids,
)
def test_has_correct_counts(summary_count, test_count):
    assert summary_count == summary_count


summary_freqs = {
    k: [d[key] for key in d if "freq" in key][0] for k, d in summaries.items()
}

test_freqs = {
    "currency": [(0, 1), (1, 3), (2, 1), (5, 1)],
    "emoji": [(0, 1), (1, 1), (2, 1)],
    "exclamation": [(0, 1), (1, 2), (2, 2)],
    "hashtag": [(0, 1), (1, 7), (2, 4)],
    "intense": [(0, 1), (1, 4), (2, 1)],
    "mention": [(0, 2), (1, 6), (2, 4)],
    "number": [(0, 1), (1, 5), (2, 1)],
    "question": [(0, 1), (1, 3), (2, 1)],
    "word_full": [(0, 4), (1, 2), (2, 1), (3, 1)],
    "word_not_full": [(0, 1), (1, 2), (2, 4), (3, 1)],
    "url": [(0, 1), (1, 2), (2, 1)],
}


@pytest.mark.parametrize(
    "summary_freq, test_freq",
    zip(sorted(summary_freqs.items()), sorted(test_freqs.items())),
    ids=test_ids,
)
def test_has_correct_freq(summary_freq, test_freq):
    assert summary_freq == summary_freq


summary_flats = {
    k: [d[key] for key in d if "flat" in key and "emoji_flat_text" not in key][0]
    for k, d in summaries.items()
}
test_flats = {
    "currency": ["$", "£", "€", "$", "£", "₣", "₤", "₥", "₦", "₧"],
    "emoji": ["😀", "😀", "😉"],
    "exclamation": ["!", "!", "!", "¡", "!", "!"],
    "hashtag": [
        "#name",
        "#oneword",
        "#nam",
        "#name",
        "#first",
        "#last",
        "#under_score",
        "#dot",
        "#مرحبا",
        "＃sign",
        "#one",
        "#two",
        "#123text",
        "#_before",
        "#after_",
    ],
    "intense": ["looooove", "goooood", "mooorning", "thhhhhhe", "innnnnn", "end!!!!!!"],
    "mention": [
        "@name",
        "@oneword",
        "@nam",
        "@name",
        "@first",
        "@last",
        "@under_score",
        "@dot",
        "＠sign",
        "@one",
        "@two",
        "@123text",
        "@_before",
        "@after_",
    ],
    "number": ["123,000", "123", "123,456", "123.234.3", "123", "123", "456,789"],
    "question": ["?", "?", "¿", "?", ";"],
    "word_full": ["rain", "snow", "rain", "snow", "rain", "snow", "rain"],
    "word_not_full": [
        "raining",
        "rain",
        "raining",
        "snowing",
        "snowing",
        "raining",
        "training",
        "snow",
        "rain",
        "snow",
        "@rain,",
        "#snow",
        "rain",
    ],
    "url": [
        "https://www.a.com",
        "http://www.a.com",
        "http://www.b.com",
        "https://example.com/one?a=b#nothing",
    ],
}


@pytest.mark.parametrize(
    "summary_flat, test_flat",
    zip(sorted(summary_flats.items()), sorted(test_flats.items())),
    ids=test_ids,
)
def test_has_correct_flat(summary_flat, test_flat):
    assert summary_flat == test_flat


summary_tops = {k: [d[key] for key in d if "top" in key] for k, d in summaries.items()}

summary_tops = {k: d[0] if len(d) == 1 else d for k, d in summary_tops.items()}

test_tops = {
    "currency": [
        ("$", 2),
        ("£", 2),
        ("€", 1),
        ("₣", 1),
        ("₤", 1),
        ("₥", 1),
        ("₦", 1),
        ("₧", 1),
    ],
    "emoji": [
        [("😀", 2), ("😉", 1)],
        [("grinning face", 2), ("winking face", 1)],
        [("Smileys & Emotion", 3)],
        [("face-smiling", 3)],
    ],
    "exclamation": [("!", 5), ("¡", 1)],
    "hashtag": [
        ("#name", 2),
        ("#oneword", 1),
        ("#nam", 1),
        ("#first", 1),
        ("#last", 1),
        ("#under_score", 1),
        ("#dot", 1),
        ("#مرحبا", 1),
        ("＃sign", 1),
        ("#one", 1),
        ("#two", 1),
        ("#123text", 1),
        ("#_before", 1),
        ("#after_", 1),
    ],
    "intense": [
        ("looooove", 1),
        ("goooood", 1),
        ("mooorning", 1),
        ("thhhhhhe", 1),
        ("innnnnn", 1),
        ("end!!!!!!", 1),
    ],
    "mention": [
        ("@name", 2),
        ("@oneword", 1),
        ("@nam", 1),
        ("@first", 1),
        ("@last", 1),
        ("@under_score", 1),
        ("@dot", 1),
        ("＠sign", 1),
        ("@one", 1),
        ("@two", 1),
        ("@123text", 1),
        ("@_before", 1),
        ("@after_", 1),
    ],
    "number": [
        ("123", 3),
        ("123,000", 1),
        ("123,456", 1),
        ("123.234.3", 1),
        ("456,789", 1),
    ],
    "question": [("?", 3), ("¿", 1), (";", 1)],
    "word_full": [("rain", 4), ("snow", 3)],
    "word_not_full": [
        ("raining", 3),
        ("rain", 3),
        ("snowing", 2),
        ("snow", 2),
        ("training", 1),
        ("@rain,", 1),
        ("#snow", 1),
    ],
    "url": [
        [
            ("https://www.a.com", 1),
            ("http://www.a.com", 1),
            ("http://www.b.com", 1),
            ("https://example.com/one?a=b#nothing", 1),
        ],
        [("www.a.com", 2), ("www.b.com", 1), ("example.com", 1)],
        [("com", 4)],
    ],
}


def make_hashable(iterable):
    """Some sub-elements are lists and need to be converted sorted tuples."""
    return tuple(tuple(sorted(x)) if isinstance(x, list) else x for x in iterable)


@pytest.mark.parametrize(
    "summary_top, test_top",
    zip(sorted(summary_tops.items()), sorted(test_tops.items())),
    ids=test_ids,
)
def test_has_correct_top(summary_top, test_top):
    assert set(make_hashable(summary_top[1])) == set(make_hashable(test_top[1]))


summary_overviews = {
    k: [d[key] for key in d if "overview" in key][0] for k, d in summaries.items()
}

test_overviews = {
    "currency": [6, 10, 10 / 6, 8],
    "emoji": [3, 3, 1.0, 2],
    "exclamation": [5, 6, 1.2, 2],
    "hashtag": [12, 15, 1.25, 14],
    "intense": [6, 6, 1.0, 6],
    "mention": [12, 14, 14 / 12, 13],
    "number": [7, 7, 7 / 7, 5],
    "question": [5, 5, 1.0, 3],
    "word_full": [8, 7, 7 / 8, 2],
    "word_not_full": [8, 13, 13 / 8, 7],
    "url": [4, 4, 1.0, 4],
}


def dict2overview_list(d):
    """Convert an overview dict to a list based on its keys."""
    result = [0, 0, 0, 0]
    for key in d:
        if "num_posts" in key:
            result[0] = d[key]
        if "num_" in key and "post" not in key:
            result[1] = d[key]
        if "per" in key:
            result[2] = d[key]
        if "unique" in key:
            result[3] = d[key]
    return result


@pytest.mark.parametrize(
    "summary_overview, test_overview",
    zip(sorted(summary_overviews.items()), sorted(test_overviews.items())),
    ids=test_ids,
)
def test_has_correct_overview(summary_overview, test_overview):
    assert dict2overview_list(summary_overview[1]) == test_overview[1]


def test_extract_puts_str_in_list():
    result = extract("#one #two #three", regex=r"#\w+", key_name="hashtag")
    assert result["hashtags"] == ["#one #two #three".split()]


def test_extract_words_puts_str_in_list():
    word_summary_str = extract_words(word_posts, "rain", True)
    assert word_summary_str["top_words"][0][0] == "rain"


def test_extract_numbers_works_without_separators():
    result = extract_numbers("123,456 hello ", number_separators=None)
    assert result["numbers"] == [["123", "456"]]


def test_extract_numbers_handles_dash_in_the_middle_of_seps():
    result = extract_numbers("123,456-789", number_separators=(".", "-", ","))
    assert result["numbers"] == [["123,456-789"]]



"""
.. _serp:

Import Search Engine Results Pages (SERPs) for Google and YouTube
=================================================================
Analyzing a single SERP is like getting one person to fill out a questionnaire
and calling it a survey.

Just like surveys, SERPs need to be collected in large-enough numbers that are
representative of the industry/market you want to understand. This is the main
feature of the ``serp_`` functions. They allow you to get the SERPs for a list
of queries, across several dimensions (like country, search type, start
position, and so on).

There are many parameters that can be used, and you can supply a list for each.
The function will get the SERPs for the *product* of all those lists. For
example, let's say you you provide the following arguments to the
:func:`serp_goog` function:

* `q`: ['serp tools', 'best serp tools', 'serp tool reviews']
* `gl`: ['us', 'ca', 'uk', 'au', 'nz']
* `start`: [1, 11, 21]

The function will produce:
3 (queries) x 5 (countries) x 3 (start positions) = 45 requests

You typically get ten results each, so in this case you would get 450 rows of
data.

All this is done in with one line of code. The result is a single DataFrame
with a row for each result, and columns for each attribute (title, snippet,
etc.), as well as meta data columns, like `queryTime` and the parameters you
selected (`q`, `gl`, and `start` in this case).


Before being able to run queries using :func:`serp_goog`, you will need to set
up some credentials as follows (you don't need a custom search engine for
:func:`serp_youtube`):

* `Create a custom search engine <https://cse.google.com/>`_: At first, you might be
  asked to enter a site to search. Enter any domain, then go to the control panel and
  remove it. Make sure you enable "Search the entire web" and image search. You will
  also need to get your search engine ID, which you can find on the control panel page.

* `Enable the custom search API <https://console.cloud.google.com/apis/library/customsearch.googleapis.com?pli=1>`_:
  The service will allow you to retrieve and display search results from your custom
  search engine programmatically. You will need to create a project for this first.

* `Create credentials for this project <https://console.developers.google.com/apis/api/customsearch.googleapis.com/credentials>`_:
  so you can get your key.

* `Enable billing for your project <https://console.cloud.google.com/billing/projects>`_
  if you want to run more than 100 queries per day. The first 100 queries are free; then
  for each additional 1,000 queries, you pay $5.


"""

__all__ = [
    "SERP_GOOG_VALID_VALS",
    "YOUTUBE_TOPIC_IDS",
    "YOUTUBE_VID_CATEGORY_IDS",
    "serp_goog",
    "serp_youtube",
    "set_logging_level",
    "youtube_channel_details",
    "youtube_video_details",
]

import datetime
import logging
from itertools import product

import pandas as pd

if int(pd.__version__[0]) >= 1:
    from pandas import json_normalize
else:
    from pandas.io.json import json_normalize

import requests

SERP_GOOG_LOG_FMT = (
    "%(asctime)s | %(levelname)s | %(filename)s:%(lineno)d "
    "| %(funcName)s | %(message)s"
)
logging.basicConfig(format=SERP_GOOG_LOG_FMT)


##############################################################################
# Google variables
##############################################################################


SERP_GOOG_VALID_VALS = dict(
    fileType={
        "bas",
        "c",
        "cc",
        "cpp",
        "cs",
        "cxx",
        "doc",
        "docx",
        "dwf",
        "gpx",
        "h",
        "hpp",
        "htm",
        "html",
        "hwp",
        "java",
        "kml",
        "kmz",
        "odp",
        "ods",
        "odt",
        "pdf",
        "pl",
        "ppt",
        "pptx",
        "ps",
        "py",
        "rtf",
        "svg",
        "swf",
        "tex",
        "text",
        "txt",
        "wap",
        "wml",
        "xls",
        "xlsx",
        "xml",
    },
    c2coff={0, 1},
    cr={
        "countryAF",
        "countryAL",
        "countryDZ",
        "countryAS",
        "countryAD",
        "countryAO",
        "countryAI",
        "countryAQ",
        "countryAG",
        "countryAR",
        "countryAM",
        "countryAW",
        "countryAU",
        "countryAT",
        "countryAZ",
        "countryBS",
        "countryBH",
        "countryBD",
        "countryBB",
        "countryBY",
        "countryBE",
        "countryBZ",
        "countryBJ",
        "countryBM",
        "countryBT",
        "countryBO",
        "countryBA",
        "countryBW",
        "countryBV",
        "countryBR",
        "countryIO",
        "countryBN",
        "countryBG",
        "countryBF",
        "countryBI",
        "countryKH",
        "countryCM",
        "countryCA",
        "countryCV",
        "countryKY",
        "countryCF",
        "countryTD",
        "countryCL",
        "countryCN",
        "countryCX",
        "countryCC",
        "countryCO",
        "countryKM",
        "countryCG",
        "countryCD",
        "countryCK",
        "countryCR",
        "countryCI",
        "countryHR",
        "countryCU",
        "countryCY",
        "countryCZ",
        "countryDK",
        "countryDJ",
        "countryDM",
        "countryDO",
        "countryTP",
        "countryEC",
        "countryEG",
        "countrySV",
        "countryGQ",
        "countryER",
        "countryEE",
        "countryET",
        "countryEU",
        "countryFK",
        "countryFO",
        "countryFJ",
        "countryFI",
        "countryFR",
        "countryFX",
        "countryGF",
        "countryPF",
        "countryTF",
        "countryGA",
        "countryGM",
        "countryGE",
        "countryDE",
        "countryGH",
        "countryGI",
        "countryGR",
        "countryGL",
        "countryGD",
        "countryGP",
        "countryGU",
        "countryGT",
        "countryGN",
        "countryGW",
        "countryGY",
        "countryHT",
        "countryHM",
        "countryVA",
        "countryHN",
        "countryHK",
        "countryHU",
        "countryIS",
        "countryIN",
        "countryID",
        "countryIR",
        "countryIQ",
        "countryIE",
        "countryIL",
        "countryIT",
        "countryJM",
        "countryJP",
        "countryJO",
        "countryKZ",
        "countryKE",
        "countryKI",
        "countryKP",
        "countryKR",
        "countryKW",
        "countryKG",
        "countryLA",
        "countryLV",
        "countryLB",
        "countryLS",
        "countryLR",
        "countryLY",
        "countryLI",
        "countryLT",
        "countryLU",
        "countryMO",
        "countryMK",
        "countryMG",
        "countryMW",
        "countryMY",
        "countryMV",
        "countryML",
        "countryMT",
        "countryMH",
        "countryMQ",
        "countryMR",
        "countryMU",
        "countryYT",
        "countryMX",
        "countryFM",
        "countryMD",
        "countryMC",
        "countryMN",
        "countryMS",
        "countryMA",
        "countryMZ",
        "countryMM",
        "countryNA",
        "countryNR",
        "countryNP",
        "countryNL",
        "countryAN",
        "countryNC",
        "countryNZ",
        "countryNI",
        "countryNE",
        "countryNG",
        "countryNU",
        "countryNF",
        "countryMP",
        "countryNO",
        "countryOM",
        "countryPK",
        "countryPW",
        "countryPS",
        "countryPA",
        "countryPG",
        "countryPY",
        "countryPE",
        "countryPH",
        "countryPN",
        "countryPL",
        "countryPT",
        "countryPR",
        "countryQA",
        "countryRE",
        "countryRO",
        "countryRU",
        "countryRW",
        "countrySH",
        "countryKN",
        "countryLC",
        "countryPM",
        "countryVC",
        "countryWS",
        "countrySM",
        "countryST",
        "countrySA",
        "countrySN",
        "countryCS",
        "countrySC",
        "countrySL",
        "countrySG",
        "countrySK",
        "countrySI",
        "countrySB",
        "countrySO",
        "countryZA",
        "countryGS",
        "countryES",
        "countryLK",
        "countrySD",
        "countrySR",
        "countrySJ",
        "countrySZ",
        "countrySE",
        "countryCH",
        "countrySY",
        "countryTW",
        "countryTJ",
        "countryTZ",
        "countryTH",
        "countryTG",
        "countryTK",
        "countryTO",
        "countryTT",
        "countryTN",
        "countryTR",
        "countryTM",
        "countryTC",
        "countryTV",
        "countryUG",
        "countryUA",
        "countryAE",
        "countryUK",
        "countryUS",
        "countryUM",
        "countryUY",
        "countryUZ",
        "countryVU",
        "countryVE",
        "countryVN",
        "countryVG",
        "countryVI",
        "countryWF",
        "countryEH",
        "countryYE",
        "countryYU",
        "countryZM",
        "countryZW",
    },
    gl={
        "ad",
        "ae",
        "af",
        "ag",
        "ai",
        "al",
        "am",
        "an",
        "ao",
        "aq",
        "ar",
        "as",
        "at",
        "au",
        "aw",
        "az",
        "ba",
        "bb",
        "bd",
        "be",
        "bf",
        "bg",
        "bh",
        "bi",
        "bj",
        "bm",
        "bn",
        "bo",
        "br",
        "bs",
        "bt",
        "bv",
        "bw",
        "by",
        "bz",
        "ca",
        "cc",
        "cd",
        "cf",
        "cg",
        "ch",
        "ci",
        "ck",
        "cl",
        "cm",
        "cn",
        "co",
        "cr",
        "cs",
        "cu",
        "cv",
        "cx",
        "cy",
        "cz",
        "de",
        "dj",
        "dk",
        "dm",
        "do",
        "dz",
        "ec",
        "ee",
        "eg",
        "eh",
        "er",
        "es",
        "et",
        "fi",
        "fj",
        "fk",
        "fm",
        "fo",
        "fr",
        "ga",
        "gd",
        "ge",
        "gf",
        "gh",
        "gi",
        "gl",
        "gm",
        "gn",
        "gp",
        "gq",
        "gr",
        "gs",
        "gt",
        "gu",
        "gw",
        "gy",
        "hk",
        "hm",
        "hn",
        "hr",
        "ht",
        "hu",
        "id",
        "ie",
        "il",
        "in",
        "io",
        "iq",
        "ir",
        "is",
        "it",
        "jm",
        "jo",
        "jp",
        "ke",
        "kg",
        "kh",
        "ki",
        "km",
        "kn",
        "kp",
        "kr",
        "kw",
        "ky",
        "kz",
        "la",
        "lb",
        "lc",
        "li",
        "lk",
        "lr",
        "ls",
        "lt",
        "lu",
        "lv",
        "ly",
        "ma",
        "mc",
        "md",
        "mg",
        "mh",
        "mk",
        "ml",
        "mm",
        "mn",
        "mo",
        "mp",
        "mq",
        "mr",
        "ms",
        "mt",
        "mu",
        "mv",
        "mw",
        "mx",
        "my",
        "mz",
        "na",
        "nc",
        "ne",
        "nf",
        "ng",
        "ni",
        "nl",
        "no",
        "np",
        "nr",
        "nu",
        "nz",
        "om",
        "pa",
        "pe",
        "pf",
        "pg",
        "ph",
        "pk",
        "pl",
        "pm",
        "pn",
        "pr",
        "ps",
        "pt",
        "pw",
        "py",
        "qa",
        "re",
        "ro",
        "ru",
        "rw",
        "sa",
        "sb",
        "sc",
        "sd",
        "se",
        "sg",
        "sh",
        "si",
        "sj",
        "sk",
        "sl",
        "sm",
        "sn",
        "so",
        "sr",
        "st",
        "sv",
        "sy",
        "sz",
        "tc",
        "td",
        "tf",
        "tg",
        "th",
        "tj",
        "tk",
        "tl",
        "tm",
        "tn",
        "to",
        "tr",
        "tt",
        "tv",
        "tw",
        "tz",
        "ua",
        "ug",
        "uk",
        "um",
        "us",
        "uy",
        "uz",
        "va",
        "vc",
        "ve",
        "vg",
        "vi",
        "vn",
        "vu",
        "wf",
        "ws",
        "ye",
        "yt",
        "za",
        "zm",
        "zw",
    },
    filter={0, 1},
    hl={
        "af",
        "sq",
        "sm",
        "ar",
        "az",
        "eu",
        "be",
        "bn",
        "bh",
        "bs",
        "bg",
        "ca",
        "zh-CN",
        "zh-TW",
        "hr",
        "cs",
        "da",
        "nl",
        "en",
        "eo",
        "et",
        "fo",
        "fi",
        "fr",
        "fy",
        "gl",
        "ka",
        "de",
        "el",
        "gu",
        "iw",
        "hi",
        "hu",
        "is",
        "id",
        "ia",
        "ga",
        "it",
        "ja",
        "jw",
        "kn",
        "ko",
        "la",
        "lv",
        "lt",
        "mk",
        "ms",
        "ml",
        "mt",
        "mr",
        "ne",
        "no",
        "nn",
        "oc",
        "fa",
        "pl",
        "pt-BR",
        "pt-PT",
        "pa",
        "ro",
        "ru",
        "gd",
        "sr",
        "si",
        "sk",
        "sl",
        "es",
        "su",
        "sw",
        "sv",
        "tl",
        "ta",
        "te",
        "th",
        "ti",
        "tr",
        "uk",
        "ur",
        "uz",
        "vi",
        "cy",
        "xh",
        "zu",
    },
    imgColorType={"color", "gray", "mono", "trans"},
    imgDominantColor={
        "black",
        "blue",
        "brown",
        "gray",
        "green",
        "orange",
        "pink",
        "purple",
        "red",
        "teal",
        "white",
        "yellow",
    },
    imgSize={
        "huge",
        "icon",
        "large",
        "medium",
        "small",
        "xlarge",
        "xxlarge",
    },
    imgType={"clipart", "face", "lineart", "stock", "photo", "animated"},
    lr={
        "lang_ar",
        "lang_bg",
        "lang_ca",
        "lang_zh-CN",
        "lang_zh-TW",
        "lang_hr",
        "lang_cs",
        "lang_da",
        "lang_nl",
        "lang_en",
        "lang_et",
        "lang_fi",
        "lang_fr",
        "lang_de",
        "lang_el",
        "lang_iw",
        "lang_hu",
        "lang_is",
        "lang_id",
        "lang_it",
        "lang_ja",
        "lang_ko",
        "lang_lv",
        "lang_lt",
        "lang_no",
        "lang_pl",
        "lang_pt",
        "lang_ro",
        "lang_ru",
        "lang_sr",
        "lang_sk",
        "lang_sl",
        "lang_es",
        "lang_sv",
        "lang_tr",
    },
    num={1, 2, 3, 4, 5, 6, 7, 8, 9, 10},
    rights={
        "cc_publicdomain",
        "cc_attribute",
        "cc_sharealike",
        "cc_noncommercial",
        "cc_nonderived",
    },
    safe={"active", "off"},
    searchType={None, "image"},
    siteSearchFilter={"e", "i"},
    start=range(1, 92),
)


##############################################################################
# YouTube variables
##############################################################################


YOUTUBE_TOPIC_IDS = {
    "Entertainment topics": {
        "Entertainment (parent topic)": "/m/02jjt",
        "Humor": "/m/09kqc",
        "Movies": "/m/02vxn",
        "Performing arts": "/m/05qjc",
        "Professional wrestling": "/m/066wd",
        "TV shows": "/m/0f2f9",
    },
    "Gaming topics": {
        "Action game": "/m/025zzc",
        "Action-adventure game": "/m/02ntfj",
        "Casual game": "/m/0b1vjn",
        "Gaming (parent topic)": "/m/0bzvm2",
        "Music video game": "/m/02hygl",
        "Puzzle video game": "/m/04q1x3q",
        "Racing video game": "/m/01sjng",
        "Role-playing video game": "/m/0403l3g",
        "Simulation video game": "/m/021bp2",
        "Sports game": "/m/022dc6",
        "Strategy video game": "/m/03hf_rm",
    },
    "Lifestyle topics": {
        "Fashion": "/m/032tl",
        "Fitness": "/m/027x7n",
        "Food": "/m/02wbm",
        "Hobby": "/m/03glg",
        "Lifestyle (parent topic)": "/m/019_rr",
        "Pets": "/m/068hy",
        "Physical attractiveness [Beauty]": "/m/041xxh",
        "Technology": "/m/07c1v",
        "Tourism": "/m/07bxq",
        "Vehicles": "/m/07yv9",
    },
    "Music topics": {
        "Christian music": "/m/02mscn",
        "Classical music": "/m/0ggq0m",
        "Country": "/m/01lyv",
        "Electronic music": "/m/02lkt",
        "Hip hop music": "/m/0glt670",
        "Independent music": "/m/05rwpb",
        "Jazz": "/m/03_d0",
        "Music (parent topic)": "/m/04rlf",
        "Music of Asia": "/m/028sqc",
        "Music of Latin America": "/m/0g293",
        "Pop music": "/m/064t9",
        "Reggae": "/m/06cqb",
        "Rhythm and blues": "/m/06j6l",
        "Rock music": "/m/06by7",
        "Soul music": "/m/0gywn",
    },
    "Other topics": {"Knowledge": "/m/01k8wb"},
    "Society topics": {
        "Business": "/m/09s1f",
        "Health": "/m/0kt51",
        "Military": "/m/01h6rj",
        "Politics": "/m/05qt0",
        "Religion": "/m/06bvp",
        "Society (parent topic)": "/m/098wr",
    },
    "Sports topics": {
        "American football": "/m/0jm_",
        "Baseball": "/m/018jz",
        "Basketball": "/m/018w8",
        "Boxing": "/m/01cgz",
        "Cricket": "/m/09xp_",
        "Football": "/m/02vx4",
        "Golf": "/m/037hz",
        "Ice hockey": "/m/03tmr",
        "Mixed martial arts": "/m/01h7lh",
        "Motorsport": "/m/0410tth",
        "Sports (parent topic)": "/m/06ntj",
        "Tennis": "/m/07bs0",
        "Volleyball": "/m/07_53",
    },
}

YOUTUBE_VID_CATEGORY_IDS = {
    "Action/Adventure": "32",
    "Anime/Animation": "31",
    "Autos & Vehicles": "2",
    "Classics": "33",
    "Comedy": "34",
    "Documentary": "35",
    "Drama": "36",
    "Education": "27",
    "Entertainment": "24",
    "Family": "37",
    "Film & Animation": "1",
    "Foreign": "38",
    "Gaming": "20",
    "Horror": "39",
    "Howto & Style": "26",
    "Movies": "30",
    "Music": "10",
    "News & Politics": "25",
    "Nonprofits & Activism": "29",
    "People & Blogs": "22",
    "Pets & Animals": "15",
    "Sci-Fi/Fantasy": "40",
    "Science & Technology": "28",
    "Short Movies": "18",
    "Shorts": "42",
    "Shows": "43",
    "Sports": "17",
    "Thriller": "41",
    "Trailers": "44",
    "Travel & Events": "19",
    "Videoblogging": "21",
}

SERP_YTUBE_VALID_VALS = dict(
    channelType={"any", "show"},
    eventType={"completed", "live", "upcoming"},
    forContentOwner={True, False, "true", "false"},
    forDeveloper={True, False, "true", "false"},
    forMine={True, False, "true", "false"},
    maxResults=range(51),
    order={"date", "rating", "relevance", "title", "videoCount", "viewCount"},
    regionCode={
        "ad",
        "ae",
        "af",
        "ag",
        "ai",
        "al",
        "am",
        "an",
        "ao",
        "aq",
        "ar",
        "as",
        "at",
        "au",
        "aw",
        "az",
        "ba",
        "bb",
        "bd",
        "be",
        "bf",
        "bg",
        "bh",
        "bi",
        "bj",
        "bm",
        "bn",
        "bo",
        "br",
        "bs",
        "bt",
        "bv",
        "bw",
        "by",
        "bz",
        "ca",
        "cc",
        "cd",
        "cf",
        "cg",
        "ch",
        "ci",
        "ck",
        "cl",
        "cm",
        "cn",
        "co",
        "cr",
        "cs",
        "cu",
        "cv",
        "cx",
        "cy",
        "cz",
        "de",
        "dj",
        "dk",
        "dm",
        "do",
        "dz",
        "ec",
        "ee",
        "eg",
        "eh",
        "er",
        "es",
        "et",
        "fi",
        "fj",
        "fk",
        "fm",
        "fo",
        "fr",
        "ga",
        "gd",
        "ge",
        "gf",
        "gh",
        "gi",
        "gl",
        "gm",
        "gn",
        "gp",
        "gq",
        "gr",
        "gs",
        "gt",
        "gu",
        "gw",
        "gy",
        "hk",
        "hm",
        "hn",
        "hr",
        "ht",
        "hu",
        "id",
        "ie",
        "il",
        "in",
        "io",
        "iq",
        "ir",
        "is",
        "it",
        "jm",
        "jo",
        "jp",
        "ke",
        "kg",
        "kh",
        "ki",
        "km",
        "kn",
        "kp",
        "kr",
        "kw",
        "ky",
        "kz",
        "la",
        "lb",
        "lc",
        "li",
        "lk",
        "lr",
        "ls",
        "lt",
        "lu",
        "lv",
        "ly",
        "ma",
        "mc",
        "md",
        "mg",
        "mh",
        "mk",
        "ml",
        "mm",
        "mn",
        "mo",
        "mp",
        "mq",
        "mr",
        "ms",
        "mt",
        "mu",
        "mv",
        "mw",
        "mx",
        "my",
        "mz",
        "na",
        "nc",
        "ne",
        "nf",
        "ng",
        "ni",
        "nl",
        "no",
        "np",
        "nr",
        "nu",
        "nz",
        "om",
        "pa",
        "pe",
        "pf",
        "pg",
        "ph",
        "pk",
        "pl",
        "pm",
        "pn",
        "pr",
        "ps",
        "pt",
        "pw",
        "py",
        "qa",
        "re",
        "ro",
        "ru",
        "rw",
        "sa",
        "sb",
        "sc",
        "sd",
        "se",
        "sg",
        "sh",
        "si",
        "sj",
        "sk",
        "sl",
        "sm",
        "sn",
        "so",
        "sr",
        "st",
        "sv",
        "sy",
        "sz",
        "tc",
        "td",
        "tf",
        "tg",
        "th",
        "tj",
        "tk",
        "tl",
        "tm",
        "tn",
        "to",
        "tr",
        "tt",
        "tv",
        "tw",
        "tz",
        "ua",
        "ug",
        "uk",
        "um",
        "us",
        "uy",
        "uz",
        "va",
        "vc",
        "ve",
        "vg",
        "vi",
        "vn",
        "vu",
        "wf",
        "ws",
        "ye",
        "yt",
        "za",
        "zm",
        "zw",
    },
    relevanceLanguage={
        "af",
        "sq",
        "sm",
        "ar",
        "az",
        "eu",
        "be",
        "bn",
        "bh",
        "bs",
        "bg",
        "ca",
        "zh-CN",
        "zh-TW",
        "zh-Hans",
        "zh-Hant",
        "hr",
        "cs",
        "da",
        "nl",
        "en",
        "eo",
        "et",
        "fo",
        "fi",
        "fr",
        "fy",
        "gl",
        "ka",
        "de",
        "el",
        "gu",
        "iw",
        "hi",
        "hu",
        "is",
        "id",
        "ia",
        "ga",
        "it",
        "ja",
        "jw",
        "kn",
        "ko",
        "la",
        "lv",
        "lt",
        "mk",
        "ms",
        "ml",
        "mt",
        "mr",
        "ne",
        "no",
        "nn",
        "oc",
        "fa",
        "pl",
        "pt-BR",
        "pt-PT",
        "pa",
        "ro",
        "ru",
        "gd",
        "sr",
        "si",
        "sk",
        "sl",
        "es",
        "su",
        "sw",
        "sv",
        "tl",
        "ta",
        "te",
        "th",
        "ti",
        "tr",
        "uk",
        "ur",
        "uz",
        "vi",
        "cy",
        "xh",
        "zu",
    },
    safeSearch={"moderate", "none", "strict"},
    topicId={
        "/m/04rlf",
        "/m/02mscn",
        "/m/0ggq0m",
        "/m/01lyv",
        "/m/02lkt",
        "/m/0glt670",
        "/m/05rwpb",
        "/m/03_d0",
        "/m/028sqc",
        "/m/0g293",
        "/m/064t9",
        "/m/06cqb",
        "/m/06j6l",
        "/m/06by7",
        "/m/0gywn",
        "/m/0bzvm2",
        "/m/025zzc",
        "/m/02ntfj",
        "/m/0b1vjn",
        "/m/02hygl",
        "/m/04q1x3q",
        "/m/01sjng",
        "/m/0403l3g",
        "/m/021bp2",
        "/m/022dc6",
        "/m/03hf_rm",
        "/m/06ntj",
        "/m/0jm_",
        "/m/018jz",
        "/m/018w8",
        "/m/01cgz",
        "/m/09xp_",
        "/m/02vx4",
        "/m/037hz",
        "/m/03tmr",
        "/m/01h7lh",
        "/m/0410tth",
        "/m/07bs0",
        "/m/07_53",
        "/m/02jjt",
        "/m/09kqc",
        "/m/02vxn",
        "/m/05qjc",
        "/m/066wd",
        "/m/0f2f9",
        "/m/019_rr",
        "/m/032tl",
        "/m/027x7n",
        "/m/02wbm",
        "/m/03glg",
        "/m/068hy",
        "/m/041xxh",
        "/m/07c1v",
        "/m/07bxq",
        "/m/07yv9",
        "/m/098wr",
        "/m/09s1f",
        "/m/0kt51",
        "/m/01h6rj",
        "/m/05qt0",
        "/m/06bvp",
        "/m/01k8wb",
    },
    type={"channel", "playlist", "video"},
    videoCaption={"any", "closedCaption", "none"},
    videoCategoryId={
        "1",
        "2",
        "10",
        "15",
        "17",
        "18",
        "19",
        "20",
        "21",
        "22",
        "23",
        "24",
        "25",
        "26",
        "27",
        "28",
        "29",
        "30",
        "31",
        "32",
        "33",
        "34",
        "35",
        "36",
        "37",
        "38",
        "39",
        "40",
        "41",
        "42",
        "43",
        "44",
    },
    videoDefinition={"any", "high", "standard"},
    videoDimension={"2d", "3d", "any"},
    videoDuration={"any", "long", "medium", "short"},
    videoEmbeddable={"any", True, "true"},
    videoLicense={"any", "creativeCommon", "youtube"},
    videoSyndicated={"any", True, "true"},
    videoType={"any", "episode", "movie"},
)


def _split_by_comma(s, length=50):
    """Group a comma-separated string into a list of at-most
    ``length``-length words each."""
    str_split = s.split(",")
    str_list = []
    for i in range(0, len(str_split) + length, length):
        temp_str = ",".join(str_split[i : i + length])
        if temp_str:
            str_list.append(temp_str)
    return str_list


def youtube_video_details(key, vid_ids):
    """Return details of videos for which the ids are given.
    Assumes ``ids`` is a comma-separated list of video ids with
    no spaces.

    Parameters
    ----------
    key : str
      Your Google Developer key.
    vid_ids : str
      A comma-separated list of video ID's, with no spaces.

    Returns
    -------
    video_df : pandas.DataFrame
    """
    base_url = (
        "https://www.googleapis.com/youtube/v3/videos?part="
        "contentDetails,id,liveStreamingDetails,localizations,player,"
        "recordingDetails,snippet,statistics,status,topicDetails"
    )
    vid_ids = _split_by_comma(vid_ids, length=50)
    final_df = pd.DataFrame()
    for vid_id in vid_ids:
        params = {"id": vid_id, "key": key}
        logging.info(msg="Requesting: " + "video details")
        video_resp = requests.get(base_url, params=params)
        if video_resp.status_code >= 400:
            raise Exception(video_resp.json())
        items_df = pd.DataFrame(video_resp.json()["items"])
        details = ["snippet", "topicDetails", "statistics", "status", "contentDetails"]
        detail_df = pd.DataFrame()
        for detail in details:
            try:
                detail_df = pd.concat(
                    [
                        detail_df,
                        pd.DataFrame([x[detail] for x in video_resp.json()["items"]]),
                    ],
                    axis=1,
                )
            except KeyError:
                continue
        temp_df = pd.concat([items_df, detail_df], axis=1)
        final_df = pd.concat([final_df, temp_df], sort=False, ignore_index=True)
    return final_df


def youtube_channel_details(key, channel_ids):
    """Return details of channels for which the ids are given.
    Assumes ``ids`` is a comma-separated list of channel ids with
    no spaces.

    Parameters
    ----------
    key : str
      Your Google Developer key.
    channel_ids : str
      A comma-separated list of channel ID's, with no spaces.

    Returns
    -------
    channel_df : pandas.DataFrame
    """
    base_url = (
        "https://www.googleapis.com/youtube/v3/channels?part="
        "snippet,contentDetails,statistics"
    )
    channel_ids = _split_by_comma(channel_ids, length=50)
    final_df = pd.DataFrame()
    for channel_id in channel_ids:
        params = {"id": channel_id, "key": key}
        logging.info(msg="Requesting: " + "channel details")
        channel_resp = requests.get(base_url, params=params)
        if channel_resp.status_code >= 400:
            raise Exception(channel_resp.json())
        items_df = pd.DataFrame(channel_resp.json()["items"])
        details = ["snippet", "statistics", "contentDetails"]
        detail_df = pd.DataFrame()
        for detail in details:
            try:
                detail_df = pd.concat(
                    [
                        detail_df,
                        pd.DataFrame([x[detail] for x in channel_resp.json()["items"]]),
                    ],
                    axis=1,
                )
            except KeyError:
                continue
        temp_df = pd.concat([items_df, detail_df], axis=1)
        final_df = pd.concat([final_df, temp_df], sort=False, ignore_index=True)
    return final_df


def _dict_product(d):
    """Return the product of all values of a dict, while
    coupling each value with its key.
    This is used to generate multiple queries out of
        possibly multiple arguments in serp_goog.

    >>> d = {"a": [1], "b": [2, 3, 4], "c": [5, 6]}
    >>> _dict_product(d)
    >>> [{'a': 1, 'b': 2, 'c': 5},
         {'a': 1, 'b': 2, 'c': 6},
         {'a': 1, 'b': 3, 'c': 5},
         {'a': 1, 'b': 3, 'c': 6},
         {'a': 1, 'b': 4, 'c': 5},
         {'a': 1, 'b': 4, 'c': 6}]
    """
    items = list(d.items())
    keys = [x[0] for x in items]
    values = [x[1] for x in items]
    dicts = []
    for prod in product(*values):
        tempdict = dict(zip(keys, prod))
        dicts.append(tempdict)
    return dicts


def serp_goog(
    q,
    cx,
    key,
    c2coff=None,
    cr=None,
    dateRestrict=None,
    exactTerms=None,
    excludeTerms=None,
    fileType=None,
    filter=None,
    gl=None,
    highRange=None,
    hl=None,
    hq=None,
    imgColorType=None,
    imgDominantColor=None,
    imgSize=None,
    imgType=None,
    linkSite=None,
    lowRange=None,
    lr=None,
    num=None,
    orTerms=None,
    rights=None,
    safe=None,
    searchType=None,
    siteSearch=None,
    siteSearchFilter=None,
    sort=None,
    start=None,
):
    """Query Google's search API and get search results in a DataFrame.

    For each parameter, you can supply single or multiple values / arguments.
        If you pass multiple arguments, all the possible combinations of
        arguments (the product) will be requested, and you will get one
        DataFrame combining all queries. See examples below.

    Parameters
    ----------
    q : str
      The search expression.
    cx : str
      The custom search engine ID to use for this request.
    key : str
      The API key of your custom search engine.
    c2coff : str
      Enables or disables Simplified and Traditional Chinese Search. The default value
      for this parameter is 0 (zero), meaning that the feature is enabled. Supported
      values are:1: Disabled0: Enabled (default)
    cr : str
      Restricts search results to documents originating in a particular country. You may
      use Boolean operators in the cr parameter's value.Google Search determines the
      country of a document by analyzing:the top- level domain (TLD) of the document's
      URLthe geographic location of the Web server's IP addressSee the Country Parameter
      Values page for a list of valid values for this parameter.
    dateRestrict : str
      Restricts results to URLs based on date.

      Supported values include:
        - d[number]: requests results from the specified number of past days.
        - w[number]: requests results from the specified number of past weeks.
        - m[number]: requests results from the specified number of past months.
        - y[number]: requests results from the specified number of past years.
    exactTerms : str
      Identifies a phrase that all documents in the search results must contain.
    excludeTerms : str
      Identifies a word or phrase that should not appear in any documents in the search
      results.
    fileType : str
      Restricts results to files of a specified extension. A list of file types
      indexable by Google can be found in Search Console Help Center.
    filter : str
      Controls turning on or off the duplicate content filter.See Automatic Filtering
      for more information about Google's search results filters. Note that host
      crowding filtering applies only to multi-site searches.By default, Google applies
      filtering to all search results to improve the quality of those results.
      Acceptable values are:  "0": Turns off duplicate content filter.  "1": Turns on
      duplicate content filter.
    gl : str
      Geolocation of end user. The gl parameter value is a two-letter country code. The
      gl parameter boosts search results whose country of origin matches the parameter
      value. See the Country Codes page for a list of valid values.Specifying a gl
      parameter value should lead to more relevant results. This is particularly true
      for international customers and, even more specifically, for customers in
      English- speaking countries other than the United States.
    highRange : str
      Specifies the ending value for a search range.Use lowRange and highRange to append
      an inclusive search range of lowRange...highRange to the query.
    hl : str
      Sets the user interface language. Explicitly setting this parameter improves the
      performance and the quality of your search results.See the Interface Languages
      section of Internationalizing Queries and Results Presentation for more
      information, and Supported Interface Languages for a list of supported languages.
    hq : str
      Appends the specified query terms to the query, as if they were combined with a
      logical AND operator.
    imgColorType : str
      Returns black and white, grayscale, or color images: mono, gray, and color.
      Acceptable values are:  "color": color  "gray": gray  "mono": mono
    imgDominantColor : str
      Returns images of a specific dominant color.  Acceptable values are:
      "black": black "blue": blue  "brown": brown  "gray": gray  "green": green
      "orange": orange  "pink": pink  "purple": purple  "red": red "teal": teal
      "white": white  "yellow": yellow
    imgSize : str
      Returns images of a specified size. Acceptable values are:  "huge": huge
      "icon": icon  "large": large  "medium": medium  "small": small  "xlarge": xlarge
      "xxlarge": xxlarge
    imgType : str
      Returns images of a type.  Acceptable values are:  "clipart": clipart
      "face": face  "lineart": lineart  "news": news  "photo": photo
    linkSite : str
      Specifies that all search results should contain a link to a particular URL
    lowRange : str
      Specifies the starting value for a search range. Use lowRange and highRange to
      append an inclusive search range of lowRange...highRange to the query.
    lr : str
      Restricts the search to documents written in a particular language
      (e.g., lr=lang_ja).  Acceptable values are:  "lang_ar": Arabic
      "lang_bg": Bulgarian  "lang_ca": Catalan  "lang_cs": Czech  "lang_da": Danish
      "lang_de": German  "lang_el": Greek  "lang_en": English  "lang_es": Spanish
      "lang_et": Estonian  "lang_fi": Finnish  "lang_fr": French  "lang_hr": Croatian
      "lang_hu": Hungarian "lang_id": Indonesian  "lang_is": Icelandic
      "lang_it": Italian  "lang_iw": Hebrew  "lang_ja": Japanese  "lang_ko": Korean
      "lang_lt": Lithuanian  "lang_lv": Latvian "lang_nl": Dutch  "lang_no": Norwegian
      "lang_pl": Polish "lang_pt": Portuguese  "lang_ro": Romanian  "lang_ru": Russian
      "lang_sk": Slovak  "lang_sl": Slovenian  "lang_sr": Serbian  "lang_sv": Swedish
      "lang_tr": Turkish  "lang_zh- CN": Chinese (Simplified)  "lang_zh-TW":
      Chinese (Traditional)
    num : int
      Number of search results to return.Valid values are integers between 1 and 10,
      inclusive.
    orTerms : str
      Provides additional search terms to check for in a document, where each document
      in the search results must contain at least one of the additional search terms.
    rights : str
      Filters based on licensing. Supported values include: cc_publicdomain,
      cc_attribute, cc_sharealike, cc_noncommercial, cc_nonderived, and combinations of
      these.
    safe : str
      Search safety level.  Acceptable values are:  "active": Enables SafeSearch
      filtering.  "off":Disables SafeSearch filtering.  (default)
    searchType : str
      Specifies the search type: image. If unspecified, results are limited to webpages.
      Acceptable values are:  "image": custom image search.
    siteSearch : str
      Specifies all search results should be pages from a given site.
    siteSearchFilter : str
      Controls whether to include or exclude results from the site named in the
      siteSearch parameter.  Acceptable values are:  "e": exclude  "i": include
    sort : str
      The sort expression to apply to the results.
    start : int
      The index of the first result to return.Valid value are integers starting 1
      (default) and the second result is 2 and so forth. For example &start=11 gives the
      second page of results with the default "num" value of 10 results per page.Note:
      No more than 100 results will ever be returned for any query with JSON API, even
      if more than 100 documents match the query, so setting (start + num) to more than
      100 will produce an error. Note that the maximum value for num is 10.

    Returns
    -------
    serp_df : pandas.DataFrame

    Examples
    --------
    The following function call will produce two queries:
    "hotel" in the USA, and "hotel" in France

    >>> serp_goog(q="hotel", gl=["us", "fr"], cx="YOUR_CX", key="YOUR_KEY")

    The below function call will prouce four queries and make four requests:

    * "fligts" in UK
    * "fligts" in Australia
    * "tickets" in UK
    * "tickets" in Australia

    'cr' here refers to 'country restrict', which focuses on content
    originating from the specified country.

    >>> serp_goog(q=['flights', 'tickets'], cr=['countryUK', 'countryAU'],
                  cx='YOUR_CX', key='YOUR_KEY')
    """
    params = locals()
    supplied_params = {k: v for k, v in params.items() if params[k] is not None}

    for p in supplied_params:
        if isinstance(supplied_params[p], (str, int)):
            supplied_params[p] = [supplied_params[p]]

    for p in supplied_params:
        if p in SERP_GOOG_VALID_VALS:
            if not set(supplied_params[p]).issubset(SERP_GOOG_VALID_VALS[p]):
                raise ValueError(
                    "Please make sure you provide a"
                    ' valid value for "{}", valid values:\n'
                    "{}".format(p, sorted(SERP_GOOG_VALID_VALS[p]))
                )
    params_list = _dict_product(supplied_params)
    base_url = "https://www.googleapis.com/customsearch/v1?"
    specified_cols = [
        "searchTerms",
        "rank",
        "title",
        "snippet",
        "displayLink",
        "link",
        "queryTime",
        "totalResults",
    ]
    responses = []
    for param in params_list:
        param_log = ", ".join([k + "=" + str(v) for k, v in param.items()])
        logging.info(msg="Requesting: " + param_log)
        resp = requests.get(base_url, params=param)
        if resp.status_code >= 400:
            raise Exception(resp.json())
        responses.append(resp)
    result_df = pd.DataFrame()
    for i, resp in enumerate(responses):
        request_metadata = resp.json()["queries"]["request"][0]
        del request_metadata["title"]
        search_info = resp.json()["searchInformation"]
        if int(search_info["totalResults"]) == 0:
            df = pd.DataFrame(columns=specified_cols, index=range(1))
            df["searchTerms"] = request_metadata["searchTerms"]
            # These keys don't appear in the response so they have to be
            # added manually
            for missing in ["lr", "num", "start", "c2coff"]:
                if missing in params_list[i]:
                    df[missing] = params_list[i][missing]
        else:
            df = pd.DataFrame(resp.json()["items"])
            df["cseName"] = resp.json()["context"]["title"]
            start_idx = request_metadata["startIndex"]
            df["rank"] = range(start_idx, start_idx + len(df))
            for missing in ["lr", "num", "start", "c2coff"]:
                if missing in params_list[i]:
                    df[missing] = params_list[i][missing]
        meta_columns = {**request_metadata, **search_info}
        df = df.assign(**meta_columns)
        df["queryTime"] = datetime.datetime.now(tz=datetime.timezone.utc)
        df["queryTime"] = pd.to_datetime(df["queryTime"])
        if "image" in df:
            img_df = json_normalize(df["image"])
            img_df.columns = ["image." + c for c in img_df.columns]
            df = pd.concat([df, img_df], axis=1)
        result_df = pd.concat([result_df, df], sort=False, ignore_index=True)
    ordered_cols = (
        list(set(params_list[i]).difference({"q", "key", "cx"})) + specified_cols
    )
    non_ordered = result_df.columns.difference(set(ordered_cols))
    final_df = result_df[ordered_cols + list(non_ordered)]
    if "pagemap" in final_df:
        pagemap_df = pd.DataFrame()
        for p in final_df["pagemap"]:
            try:
                temp_pagemap_df = json_normalize(p)
                pagemap_df = pd.concat([pagemap_df, temp_pagemap_df], sort=False)
            except Exception:
                temp_pagemap_df = pd.DataFrame({"delete_me": None}, index=range(1))
                pagemap_df = pd.concat([pagemap_df, temp_pagemap_df], sort=False)
        pagemap_df = pagemap_df.reset_index(drop=True)
        if "delete_me" in pagemap_df:
            del pagemap_df["delete_me"]
        for col in pagemap_df:
            if col in final_df:
                pagemap_df = pagemap_df.rename(columns={col: "pagemap_" + col})
        final_df = pd.concat([final_df, pagemap_df], axis=1)

        if "metatags" in pagemap_df:
            metatag_df = pd.DataFrame()
            for m in pagemap_df["metatags"]:
                try:
                    temp_metatags_df = json_normalize(m)
                    metatag_df = pd.concat([metatag_df, temp_metatags_df], sort=False)
                except Exception:
                    temp_metatags_df = pd.DataFrame({"delete_me": None}, index=range(1))
                    metatag_df = pd.concat([metatag_df, temp_metatags_df], sort=False)
            metatag_df = metatag_df.reset_index(drop=True)
            if "delete_me" in metatag_df:
                del metatag_df["delete_me"]
            for col in metatag_df:
                if col in final_df:
                    metatag_df = metatag_df.rename(columns={col: "metatag_" + col})

            final_df = pd.concat([final_df, metatag_df], axis=1)
    return final_df


def serp_youtube(
    key,
    q=None,
    channelId=None,
    channelType=None,
    eventType=None,
    forContentOwner=None,
    forDeveloper=None,
    forMine=None,
    location=None,
    locationRadius=None,
    maxResults=None,
    onBehalfOfContentOwner=None,
    order=None,
    pageToken=None,
    publishedAfter=None,
    publishedBefore=None,
    regionCode=None,
    relatedToVideoId=None,
    relevanceLanguage=None,
    safeSearch=None,
    topicId=None,
    type=None,
    videoCaption=None,
    videoCategoryId=None,
    videoDefinition=None,
    videoDimension=None,
    videoDuration=None,
    videoEmbeddable=None,
    videoLicense=None,
    videoSyndicated=None,
    videoType=None,
):
    """Query the YouTube API and get search results in a DataFrame.

    For each parameter you can supply a single or multiple value(s).
    Looping and merging results is handled automatically in case of multiple
    values.

    Parameters
    ----------
    q : str
      The ``q`` parameter specifies the query term to search for. Your request can also
      use the Boolean NOT (-) and OR (|) operators to exclude videos or to find videos
      that are associated with one of several search terms. For example, to search for
      videos matching either "boating" or "sailing", set the ``q`` parameter value to
      boating|sailing. Similarly, to search for videos matching either "boating" or
      "sailing" but not "fishing", set the q parameter value to
      boating|sailing -fishing. Note that the pipe character must be URL- escaped when
      it is sent in your API request. The URL-escaped value for the pipe character is
      %7C.
    channelId : str
      The ``channelId`` parameter indicates that the API response should only contain
      resources created by the channel. Note: Search results are constrained to a
      maximum of 500 videos if your request specifies a value for the ``channelId``
      parameter and sets the ``type`` parameter value to video, but it does not also set
      one of the ``forContentOwner``, ``forDeveloper``, or ``forMine`` filters.
    channelType : str
      The ``channelType`` parameter lets you restrict a search to a particular type of
      channel. Acceptable values are:

        any - Return all channels.

        show - Only retrieve shows.
    eventType : str
      The ``eventType`` parameter restricts a search to broadcast events. If you specify
      a value for this parameter, you must also set the type parameter's value to video.
      Acceptable values are:

        completed - Only include completed broadcasts.

        live - Only include active broadcasts.

        upcoming - Only include upcoming broadcasts.

    forContentOwner : bool
      This parameter can only be used in a properly authorized request, and it is
      intended exclusively for YouTube content partners. The ``forContentOwner``
      parameter restricts the search to only retrieve videos owned by the content owner
      identified by the ``onBehalfOfContentOwner`` parameter. If ``forContentOwner`` is
      set to true, the request must also meet these requirements: The
      ``onBehalfOfContentOwner`` parameter is required.The user authorizing the request
      must be using an account linked to the specified content owner. The ``type``
      parameter value must be set to video.None of the following other parameters can be
      set: ``videoDefinition``, ``videoDimension``, ``videoDuration``, ``videoLicense``,
      ``videoEmbeddable``, ``videoSyndicated``, ``videoType``.
    forDeveloper : bool
      This parameter can only be used in a properly authorized request. The
      ``forDeveloper`` parameter restricts the search to only retrieve videos uploaded
      via the developer's application or website. The API server uses the request's
      authorization credentials to identify the developer. The ``forDeveloper``
      parameter can be used in conjunction with optional search parameters like the
      ``q`` parameter. For this feature, each uploaded video is automatically tagged
      with the project number that is associated with the developer's application in the
      Google Developers Console. When a search request subsequently sets the
      ``forDeveloper`` parameter to ``true`` the API server uses the request's
      authorization credentials to identify the developer. Therefore, a developer can
      restrict results to videos uploaded through the developer's own app or website but
      not to videos uploaded through other apps or sites.
    forMine : bool
      This parameter can only be used in a properly authorized request. The ``forMine``
      parameter restricts the search to only retrieve videos owned by the authenticated
      user. If you set this parameter to ``true``, then the ``type`` parameter's value
      must also be set to ``video``. In addition, none of the following other parameters
      can be set in the same request: ``videoDefinition``, ``videoDimension``,
      ``videoDuration``, ``videoLicense``, ``videoEmbeddable``, ``videoSyndicated``,
      ``videoType``.
    relatedToVideoId: str
      The ``relatedToVideoId`` parameter retrieves a list of videos that are related to
      the video that the parameter ``value`` identifies. The parameter ``value`` must be
      set to a YouTube video ID and, if you are using this parameter, the ``type``
      parameter must be set to video.Note that if the ``relatedToVideoId`` parameter is
      set, the only other supported parameters are ``part``, ``maxResults``,
      ``pageToken``, ``regionCode``, ``relevanceLanguage``, ``safeSearch``, ``type``
      (which must be set to video), and ``fields``.
    location : str
      The ``location`` parameter, in conjunction with the ``locationRadius`` parameter,
      defines a circular geographic area and also restricts a search to videos that
      specify, in their metadata, a geographic location that falls within that area. The
      parameter value is a string that specifies latitude/longitude coordinates e.g.
      (37.42307,-122.08427).The location parameter value identifies the point at the
      center of the area. The ``locationRadius`` parameter specifies the maximum
      distance that the location associated with a video can be from that point for the
      video to still be included in the search results. The API returns an error if your
      request specifies a value for the ``location`` parameter but does not also specify
      a value for the ``locationRadius`` parameter.
    locationRadius : str
      The ``locationRadius`` parameter, in conjunction with the ``location`` parameter,
      defines a circular geographic area. The parameter value must be a floating point
      number followed by a measurement unit. Valid measurement units are m, km, ft, and
      mi. For example, valid parameter values include 1500m, 5km, 10000ft, and 0.75mi.
      The API does not support ``locationRadius`` parameter values larger than 1000
      kilometers. Note: See the definition of the ``location`` parameter for more
      information.
    maxResults : int
      The ``maxResults`` parameter specifies the maximum number of items that should be
      returned in the result set. Acceptable values are 0 to 50, inclusive. The default
      value is 5.
    onBehalfOfContentOwner : str
      This parameter can only be used in a properly authorized request. Note: This
      parameter is intended exclusively for YouTube content partners.The
      ``onBehalfOfContentOwner`` parameter indicates that the request's authorization
      credentials identify a YouTube CMS user who is acting on behalf of the content
      owner specified in the parameter value. This parameter is intended for YouTube
      content partners that own and manage many different YouTube channels. It allows
      content owners to authenticate once and get access to all their video and channel
      data, without having to provide authentication credentials for each individual
      channel. The CMS account that the user authenticates with must be linked to the
      specified YouTube content owner.
    order : str
      The order parameter specifies the method that will be used to order resources in
      the API response. The default value is relevance. Acceptable values are:

        date - Resources are sorted in reverse chronological order based on the
        date they were created.

        rating - Resources are sorted from highest to lowest rating.

        relevance - Resources are sorted based on their relevance to the search
        query. This is the default value for this parameter.

        title - Resources are sorted alphabetically by title.

        videoCount - Channels are sorted in descending order of their number of
        uploaded videos.

        viewCount - Resources sorted from highest to lowest number of views.

        For live broadcasts, videos are sorted by number of concurrent viewers
        while the broadcasts are ongoing.
    pageToken : str
      The ``pageToken`` parameter identifies a specific page in the result set that
      should be returned. In an API response, the ``nextPageToken`` and
      ``prevPageToken`` properties identify other pages that could be retrieved.
    publishedAfter : datetime
      The ``publishedAfter`` parameter indicates that the API response should only
      contain resources created at or after the specified time. The value is an RFC 3339
      formatted date-time value (1970-01-01T00:00:00Z).
    publishedBefore : datetime
      The ``publishedBefore`` parameter indicates that the API response should only
      contain resources created before or at the specified time. The value is an RFC
      3339 formatted date-time value (1970-01-01T00:00:00Z).
    regionCode : str
      The ``regionCode`` parameter instructs the API to return search results for videos
      that can be viewed in the specified country. The parameter value is an ISO 3166-1
      alpha-2 country code.
    relevanceLanguage : str
      The ``relevanceLanguage`` parameter instructs the API to return search results
      that are most relevant to the specified language. The parameter value is typically
      an ISO 639-1 two-letter language code. However, you should use the values zh-Hans
      for simplified Chinese and zh-Hant for traditional Chinese. Please note that
      results in other languages will still be returned if they are highly relevant to
      the search query term.
    safeSearch : str
      The ``safeSearch`` parameter indicates whether the search results should include
      restricted content as well as standard content. Acceptable values are:

        moderate - YouTube will filter some content from search results and,
        at the least, will filter content that is restricted in your locale.
        Based on their content, search results could be removed from search
        results or demoted in search results. This is the default parameter
        value.

        none - YouTube will not filter the search result set.

        strict - YouTube will try to exclude all restricted content from the
        search result set.

        Based on their content, search results
        could be removed from search results or demoted in search
        results.
    topicId : str
      The ``topicId`` parameter indicates that the API response should only contain
      resources associated with the specified topic. The value identifies a Freebase
      topic ID.
    type : str
      The ``type`` parameter restricts a search query to only retrieve a particular type
      of resource. The value is a comma-separated list of resource types. The default
      value is video,channel,playlist. Acceptable values are: channel, playlist, and
      video.
    videoCaption : str
      The ``videoCaption`` parameter indicates whether the API should filter video
      search results based on whether they have captions. If you specify a value for
      this parameter, you must also set the ``type`` parameter's value to video.
      Acceptable values are:

        any - Do not filter results based on caption availability.

        closedCaption - Only include videos that have captions.

        none - Only include videos that do not have captions.
    videoCategoryId : str
      The ``videoCategoryId`` parameter filters video search results based on their
      category. If you specify a value for this parameter, you must also set the
      ``type`` parameter's value to video.
    videoDefinition : str
      The ``videoDefinition`` parameter lets you restrict a search to only include
      either high definition (HD) or standard definition (SD) videos. HD videos are
      available for playback in at least 720p, though higher resolutions, like 1080p,
      might also be available. If you specify a value for this parameter, you must also
      set the ``type`` parameter's value to video. Acceptable values are:

        any - Return all videos, regardless of their resolution.

        high - Only retrieve HD videos.

        standard - Only retrieve videos in standard definition.
    videoDimension : str
      The ``videoDimension`` parameter lets you restrict a search to only retrieve 2D or
      3D videos. If you specify a value for this parameter, you must also set the
      ``type`` parameter's value to video. Acceptable values are:

        2d - Restrict search results to exclude 3D videos.

        3d - Restrict search results to only include 3D videos.

        any - Include both 3D and non-3D videos in returned results. This is the default
        value.
    videoDuration : str
      The ``videoDuration`` parameter filters video search results based on their
      duration. If you specify a value for this parameter, you must also set the
      ``type`` parameter's value to video. Acceptable values are:

        any - Do not filter video search results based on their duration.
        This is the default value.

        long - Only include videos longer than 20 minutes.

        medium - Only include videos that are between four and 20 minutes
        long (inclusive).

        short - Only include videos that are less than four minutes long.
    videoEmbeddable : str
      The ``videoEmbeddable`` parameter lets you to restrict a search to only videos
      that can be embedded into a webpage. If you specify a value for this parameter,
      you must also set the ``type`` parameter's value to video. Acceptable values are:

        any - Return all videos, embeddable or not.

        true - Only retrieve embeddable videos.
    videoLicense : str
      The ``videoLicense`` parameter filters search results to only include videos with
      a particular license. YouTube lets video uploaders choose to attach either the
      Creative Commons license or the standard YouTube license to each of their videos.
      If you specify a value for this parameter, you must also set the ``type``
      parameter's value to video. Acceptable values are:

        any - Return all videos, regardless of which license they have,
        that match the query parameters.

        creativeCommon - Only return videos that have a Creative Commons
        license.
        Users can reuse videos with this license in other videos that they
        create.

        youtube - Only return videos that have the standard YouTube license.
    videoSyndicated : str
      The ``videoSyndicated`` parameter lets you to restrict a search to only videos
      that can be played outside youtube.com. If you specify a value for this parameter,
      you must also set the ``type`` parameter's value to video. Acceptable values are:

        any - Return all videos, syndicated or not.

        true - Only retrieve syndicated videos.
    videoType : str
      The ``videoType`` parameter lets you restrict a search to a particular type of
      videos. If you specify a value for this parameter, you must also set the ``type``
      parameter's value to video. Acceptable values are:

        any - Return all videos.

        episode - Only retrieve episodes of shows.

        movie - Only retrieve movies.
    Returns
    -------
    serp_df : pandas.DataFrame
    """
    params = locals()
    supplied_params = {k: v for k, v in params.items() if params[k]}

    type_vid_params = {
        "eventType",
        "relatedToVideoId",
        "videoCaption",
        "videoCategoryId",
        "videoDefinition",
        "videoDimension",
        "videoDuration",
        "videoEmbeddable",
        "videoLicense",
        "videoSyndicated",
        "videoType",
        "forMine",
        "forContentOwner",
    }

    if supplied_params.get("type") != "video" and type_vid_params.intersection(
        set(supplied_params.keys())
    ):
        raise ValueError(
            'You need to set type="video" if you want to set'
            " any of the following:" + str(type_vid_params)
        )

    for p in supplied_params:
        if isinstance(supplied_params[p], (str, int)):
            supplied_params[p] = [supplied_params[p]]

    for p in supplied_params:
        if p in SERP_YTUBE_VALID_VALS:
            if not set(supplied_params[p]).issubset(SERP_YTUBE_VALID_VALS[p]):
                raise ValueError(
                    "Please make sure you provide a"
                    ' valid value for "{}", valid values:\n{}'.format(
                        p, sorted([str(x) for x in SERP_YTUBE_VALID_VALS[p]])
                    )
                )

    params_list = _dict_product(supplied_params)
    base_url = "https://www.googleapis.com/youtube/v3/search?part=snippet"

    responses = []
    for param in params_list:
        param_log = ", ".join([k + "=" + str(v) for k, v in param.items()])
        logging.info(msg="Requesting: " + param_log)
        resp = requests.get(base_url, params=param)
        if resp.status_code >= 400:
            raise Exception(resp.json())
        responses.append(resp)

    result_df = pd.DataFrame()
    for i, resp in enumerate(responses):
        snippet_df = pd.DataFrame([x["snippet"] for x in resp.json()["items"]])
        id_df = pd.DataFrame([x["id"] for x in resp.json()["items"]])
        if "channelId" in id_df:
            id_df = id_df.drop("channelId", axis=1)

        if "thumbnails" in snippet_df:
            thumb_df = json_normalize(snippet_df["thumbnails"])
        else:
            thumb_df = pd.DataFrame()
        page_info = resp.json()["pageInfo"]
        temp_df = pd.concat([snippet_df, id_df, thumb_df], axis=1).assign(**page_info)
        temp_df["rank"] = range(1, len(temp_df) + 1)

        if len(temp_df) == 0:
            empty_df_cols = [
                "title",
                "description",
                "publishedAt",
                "channelTitle",
                "kind",
                "videoId",
                "channelId",
            ]
            temp_df = temp_df.assign(q=[params_list[i]["q"]])
            temp_df = temp_df.assign(**dict.fromkeys(empty_df_cols))

            temp_df = temp_df.assign(**page_info)
        del params_list[i]["key"]
        temp_df = temp_df.assign(**params_list[i])
        temp_df["nextPageToken"] = resp.json().get("nextPageToken")
        result_df = pd.concat([result_df, temp_df], sort=False, ignore_index=True)

    result_df["queryTime"] = datetime.datetime.now(tz=datetime.timezone.utc)
    result_df["queryTime"] = pd.to_datetime(result_df["queryTime"])

    specified_cols = [
        "queryTime",
        "rank",
        "title",
        "description",
        "publishedAt",
        "channelTitle",
        "totalResults",
        "kind",
    ]
    ordered_cols = list(params_list[i].keys()) + specified_cols
    non_ordered = result_df.columns.difference(set(ordered_cols))
    final_df = result_df[ordered_cols + list(non_ordered)]

    vid_ids = ",".join(final_df["videoId"].dropna())
    if vid_ids:
        vid_details_df = youtube_video_details(vid_ids=vid_ids, key=key)
        vid_details_df.columns = ["video." + x for x in vid_details_df.columns]
        final_df = pd.merge(
            final_df, vid_details_df, how="left", left_on="videoId", right_on="video.id"
        )

    channel_ids = ",".join(final_df["channelId"].dropna())
    if channel_ids:
        channel_details_df = youtube_channel_details(channel_ids=channel_ids, key=key)
        channel_details_df.columns = [
            "channel." + x for x in channel_details_df.columns
        ]

        final_df = pd.merge(
            final_df,
            channel_details_df,
            how="left",
            left_on="channelId",
            right_on="channel.id",
        )
    final_df = final_df.drop_duplicates(subset=["videoId"])
    return final_df.reset_index(drop=True)


def set_logging_level(level_or_name):
    """Change the logging level during the session.
    Acceptable values are [0, 10, 20, 30, 40, 50,
    'NOTSET', 'DEBUG', 'INFO', 'WARNING', 'ERROR',
    'CRITICAL']
    """
    lvl_names_values = [
        0,
        10,
        20,
        30,
        40,
        50,
        "NOTSET",
        "DEBUG",
        "INFO",
        "WARNING",
        "ERROR",
        "CRITICAL",
    ]
    if level_or_name not in lvl_names_values:
        raise ValueError(
            "Please make sure you supply" " a value from: {}".format(lvl_names_values)
        )
    logging.getLogger().setLevel(level_or_name)


logging.getLogger().setLevel("INFO")



import logging
import os
from itertools import product

import pandas as pd
import pytest

from advertools.serp import (  # YOUTUBE_TOPIC_IDS,; YOUTUBE_VID_CATEGORY_IDS,
    SERP_GOOG_VALID_VALS,
    SERP_YTUBE_VALID_VALS,
    _dict_product,
    serp_goog,
    serp_youtube,
    set_logging_level,
    youtube_channel_details,
    youtube_video_details,
)

goog_cse_cx = os.environ.get("GOOG_CSE_CX")
goog_cse_key = os.environ.get("GOOG_CSE_KEY")
youtube_key = os.environ.get("GOOG_CSE_KEY")

skip_api_tests = pytest.mark.skipif(
    os.environ.get("ADV_TEST_OFFLINE"), reason="Run all except API dependents"
)


def test_dict_product_produces_correct_result():
    d = {"a": [1, 2, 3], "b": [4, 5, 6], "c": [10, 20]}
    dp_exp = [
        {"a": 1, "b": 4, "c": 10},
        {"a": 1, "b": 4, "c": 20},
        {"a": 1, "b": 5, "c": 10},
        {"a": 1, "b": 5, "c": 20},
        {"a": 1, "b": 6, "c": 10},
        {"a": 1, "b": 6, "c": 20},
        {"a": 2, "b": 4, "c": 10},
        {"a": 2, "b": 4, "c": 20},
        {"a": 2, "b": 5, "c": 10},
        {"a": 2, "b": 5, "c": 20},
        {"a": 2, "b": 6, "c": 10},
        {"a": 2, "b": 6, "c": 20},
        {"a": 3, "b": 4, "c": 10},
        {"a": 3, "b": 4, "c": 20},
        {"a": 3, "b": 5, "c": 10},
        {"a": 3, "b": 5, "c": 20},
        {"a": 3, "b": 6, "c": 10},
        {"a": 3, "b": 6, "c": 20},
    ]
    dp_res = _dict_product(d)
    for d_res in dp_res:
        assert d_res in dp_exp


def test_dict_product_return_correct_types():
    d = {"a": [1], "b": [10, 20, 30], "c": (4, 5, 6)}
    dp = _dict_product(d)
    assert isinstance(dp, list)
    assert [isinstance(x, dict) for x in dp]
    assert len(dp) == len(list(product(*d.values())))


# Google search tests:
def test_serp_goog_raises_error_on_invalid_args():
    with pytest.raises(ValueError):
        for val in SERP_GOOG_VALID_VALS:
            params = {val: "WRONG VALUE"}
            serp_goog(q="q", cx="cx", key="key", **params)


@skip_api_tests
def test_serp_goog_return_correct_result():
    result = serp_goog(
        q="testing hotels", cx=goog_cse_cx, key=goog_cse_key, searchType=["image", None]
    )
    assert isinstance(result, pd.core.frame.DataFrame)
    assert "title" in result
    assert "image" in result
    assert len(result) == 20


@skip_api_tests
def test_serp_goog_handles_no_search_results():
    q = "aquerythatdoesntgetrezultssss"
    result = serp_goog(
        q=q, cx=goog_cse_cx, key=goog_cse_key, cr="countryRU", hl="zh-TW", gl="nf"
    )
    assert len(result) == 1
    assert result["searchTerms"].values[0] == q


@skip_api_tests
def test_serp_raises_error_on_wrong_key():
    with pytest.raises(Exception):  # noqa: B017
        serp_goog(q="test credit cart", cx=goog_cse_cx, key="wrong key")


@skip_api_tests
def test_serp_goog_restult_df_contains_all_provided_params():
    keys_vals_to_test = {
        k: list(SERP_GOOG_VALID_VALS[k])[0] for k in SERP_GOOG_VALID_VALS
    }
    for key, val in keys_vals_to_test.items():
        result = serp_goog(cx=goog_cse_cx, key=goog_cse_key, q="fashion", **{key: val})
        if key == "searchType":
            continue
        assert key in result.columns
        if key == "filter":
            val = str(val)
        assert result[key].iloc[0] == val


# YouTube search tests:
def test_serp_youtube_raises_error_on_invalid_args():
    with pytest.raises(ValueError):
        for val in SERP_YTUBE_VALID_VALS:
            params = {val: "WRONG VALUE"}
            serp_youtube(q="q", key="key", **params)


@skip_api_tests
def test_serp_youtube_return_correct_result():
    result = serp_youtube(
        q=["testing hotels", "testing computers"], key=youtube_key, order="date"
    )
    assert isinstance(result, pd.core.frame.DataFrame)
    assert "title" in result
    assert "rank" in result
    assert len(result) <= 10


@skip_api_tests
def test_serp_youtube_handles_no_search_results():
    q = "aquerythatdoesntgetrezultssss"
    result = serp_youtube(
        q=q,
        key=youtube_key,
        relevanceLanguage="ar",
        channelId="UCN1Q_Z0QgTKMimCv6G0IkKQ",
    )
    assert len(result) == 1
    assert result["q"].values[0] == q
    assert result["rank"].isna().values[0]


@skip_api_tests
def test_serp_youtube_raises_type_video_error():
    with pytest.raises(Exception):  # noqa: B017
        serp_youtube(key=youtube_key, videoEmbeddable=True)


@skip_api_tests
def test_serp_youtube_raises_response_error():
    with pytest.raises(Exception):  # noqa: B017
        serp_youtube(key=youtube_key, publishedAfter="wrong date fmt")


def test_correctly_changing_log_levels():
    lvl_names_values = [0, 10, 20, 30, 40, 50]
    for level in lvl_names_values:
        set_logging_level(level)
        assert logging.getLogger().level == level
    with pytest.raises(ValueError):
        set_logging_level("WRONG VALUE")


@skip_api_tests
def test_youtube_video_details_raises_error():
    with pytest.raises(Exception):  # noqa: B017
        youtube_video_details(key="WRONG KEY", vid_ids="wrong ID")


@skip_api_tests
def test_youtube_channel_details_raises_error():
    with pytest.raises(Exception):  # noqa: B017
        youtube_channel_details(key="WRONG KEY", channel_ids="wrong ID")
